from yambopy import *
from qepy import *
from schedulerpy import *
import os
import subprocess
from glob import glob
from copy import deepcopy
import matplotlib.pyplot as plt

class YamboDG_Optimize():
    """ 
    Class to generate and run convergence tests for the yambo double grid. 
    
    ** This class is useful but complex, read the description well AND/OR check its tutorial! **

    - Needs a quantum espresso scf save folder
    - Needs nscf (qe) and yambo [desired calculation type] inputs PASSED AS PwIn AND YamboIn OBJECTS
         -- If converging the double grids, an independent-particles (ip) yambo input is required 
    - Needs lists of coarse grids (CG) and fine grids (FG); [NOTE] Only random FG presently implemented. 
    - Additional arguments: directory paths, file names, experimental laser energy [for absorption], etc.
    - The workflow is divided into FOUR STEPS that can be executed separately:
        1. nscf CG [STEPS='1']
        2. nscf FG and yambo CG [STEPS='2']
        3. yambo FG [STEPS='3']
        4. if 'converge_DG' is on (therefore with yambo--> ip):
             -- TODO: Analyis, report, plot results and give ip-converged value [STEPS='4']
             -- TODO: Move the double grid generation functions to a different submodule       
 
    - Scheme of the workflow:
        -- If job submissions are used, the workflow is better submitted in subsequent steps
        -- If planning a parallel traversal (each independent branch simultaneously) of this tree 
           with job submissions, see function branch_wise_flow
           
            NSCF                                       YAMBO
            |                                          |
    step 1  CG_1              CG_2 ... CG_N            |
            |                 |                        |
    step 2  FG_11 ... FG_M1   FG_12 ... FG_M2 ...      CG_1              CG_2 ... CG_N
                                                       |                 |
    step 3                                             FG_11 ... FG_M1   FG_12 ... FG_M2 ...
                                                        \         |      |         /
                                                         \        \      |        /
    step 4                                                 _________ PLOTS ______
    
    Some optional variables
      
      - E_laser: external laser energy (for RT checks)
      - STEPS: which workflow steps to execute
      - RUN: if False, no job is submitted
      - converge_DG: if True, enables automatic double grid convergence; requires IP yambo input.
      - nscf_out, y_out_dir: pw and yambo output directories
      - qe_scheduler,y_scheduler: SchedulerPy object for cluster submission (default: bash)
      - wait_all: if cluster submission is on, forces the master python process to wait for all sumbitted jobs to complete before exiting
      - yambo_calc_type: name the yambo calculations
      - yambo_exec: either yambo, yambo_ph or yambo_rt
      - save_type: simple, elph, expanded_elph, fixsymm, fixsymm+elph
            -- NOTE: if *elph is used: prepare a symlink to elph_dir in RUN_path
      - noyambo: if RUN is True, run only QE steps plus SAVEs, but not actual yambo calculations

    Example of use (frontend):           
        .. code-block:: python
    
            YamboDG_Optimize(cg_grids,fg_grids,prefix,scf_path,pseudo_path,...,STEPS='all')
            
    Example of use (job submission: each step dependent on the one before)
    
        .. code-block:: python 
            
            scheduler1 = Scheduler(...)
            scheduler2 = Scheduler(...,dependency=scheduler1)
            scheduler3 = Scheduler(...,dependency=scheduler2)
            
            scheduler1.add_command('python dg_script.py -steps 1')
            sheduler1.run()
            
            scheduler2.add_command('python dg_script.py -steps 2')
            sheduler2.run()
            
            scheduler3.add_command('python dg_script.py -steps 3')
            sheduler3.run()            
       
        ..
        
        .. code-block:: dg_script.py 
           
           YamboDG_Optimize(cg_grids,fg_grids,prefix,scf_path,pseudo_path,...,wait_all==True,STEPS=args.steps)
    
    TO DO:
      - Separate double grid generation and double grid convergence (simple option 'converge_DG' might suffice)
      - If automatic DG convergence assessment is on, then implement MOMENTA of the abs spectra as a method to check convergence
      - Change check_nscf_completed into common/check_qe_completed

    """
    
    def __init__(self,cg_grids,fg_grids,prefix,nscf_input,ya_input,E_laser=0.,STEPS='all',RUN=True, converge_DG=False,\
                 scf_save_path='./scf',pseudo_path='./pseudos',RUN_path='./',nscf_out="nscf",y_out_dir="results",\
                 qe_scheduler=None,y_scheduler=None,wait_all=True,pw_exec_path='',yambo_exec_path='',\
                 yambo_exec='yambo',save_type='simple',yambo_calc_type="yambo",noyambo=False):

        #Configuring schedulers
        self.frontend = Scheduler.factory(scheduler="bash")
        if y_scheduler is not None and qe_scheduler is not None: #Here we use, e.g., slurm 
            self.qejobrun, self.yjobrun = qe_scheduler, y_scheduler
            self.wait_up = True #This will enable calls to a function that will make the code wait upon completion of previous submitted job
            self.job_folders = []
            self.job_shells  = []
        elif y_scheduler is None and qe_scheduler is None: # Both schedulers must be present to activate job submission
            self.qejobrun, self.yjobrun = Scheduler.factory(scheduler="bash"), Scheduler.factory(scheduler="bash")
            self.wait_up = False
        else: raise UserWarning('Submission mode is on only for either yambo or qe')

        #Setting global variables
        self.cg_grids = cg_grids
        self.cg_strings = []
        for cg in cg_grids: self.cg_strings.append("%dx%dx%d"%(cg[0],cg[1],cg[2]))
        self.fg_grids = fg_grids
        self.fg_strings = []
        for fg in self.fg_grids:
                temp_ls = []
                for i in range(len(fg)): temp_ls.append(str(fg[i])+'_fg')
                self.fg_strings.append(temp_ls)
        self.prefix = prefix
        self.scf_save_path = scf_save_path
        self.pseudo_path = pseudo_path
        self.RUN_path = RUN_path
        self.yambo_calc_type = yambo_calc_type
        self.noyambo = noyambo
        self.E_laser = E_laser
        
        # Initialize JOBID lists (used only in submission mode)
        self.qe_id_cg = [ -1 for cg in self.cg_strings ]
        self.ya_id_cg = [ -1 for cg in self.cg_strings ]
        self.qe_id_fg = [ [ -1 for fg_i in fg ] for fg in self.fg_strings ]
        self.ya_id_fg = [ [ -1 for fg_i in fg ] for fg in self.fg_strings ]
        
        # Path of nscf and ip calculations and final plots
        if converge_DG: self.yambo_calc_type='ip'
        if not os.path.isdir(RUN_path): os.mkdir(RUN_path)
        self.nscf_dir = '%s/nscf_grids'%RUN_path
        self.yambo_dir = '%s/%s_grids'%(RUN_path,self.yambo_calc_type)
        self.plot_dir = '%s/plots'%RUN_path
        if not os.path.isdir(self.nscf_dir): os.mkdir(self.nscf_dir)
        if not os.path.isdir(self.yambo_dir): os.mkdir(self.yambo_dir)

        #Inputs
        self.nscf_inp = nscf_input
        self.ya_inp   = ya_input
        if converge_DG:
            yambo_exec = 'yambo'
            self.ip_input_is_there()
        
        #Executables
        if yambo_exec_path != '': yambo_exec_path_aux=yambo_exec_path+'/'
        if pw_exec_path != '': pw_exec_path+='/'
        self.pw = pw_exec_path + 'pw.x'
        self.ypp = yambo_exec_path_aux + 'ypp'
        self.yambo = yambo_exec_path_aux + yambo_exec
        
        # Automatically determine which SAVE to create (better to specify it explicitly)
        if yambo_exec == 'yambo': save_type='simple'
        elif yambo_exec == 'yambo_ph' and not save_type[-4:]=='elph': save_type='elph'
        elif yambo_exec == 'yambo_rt' and not save_type[3:]=='fix': save_type='fixsymm'
        # Deal with elph_path
        elph_path = None
        if save_type[-4:]=='elph': 
            if ( not os.path.isdir('%s/elph_dir'%self.RUN_path) ) and ( not os.path.isfile('%s/elph_dir'%self.RUN_path) ):
                raise FileNotFoundError('Please mv or symlink the elph_dir folder to the RUN_path of this workflow.')
            else:
                elph_path=self.RUN_path

        #Start IO
        self.yf = YamboIO(out_name='YAMBOPY_double-grid_Optimize.log',out_path=self.RUN_path,print_to_shell=True)
        self.yf.IO_start()
        
        if converge_DG: self.yf.msg('#### DOUBLE GRID CONVERGENCE WORKFLOW ####')
        else: self.yf.msg('#### DOUBLE GRID WORKFLOW FOR %s CALCULATIONS ####'%self.yambo_calc_type)
        
        self.driver(STEPS,RUN,nscf_out,y_out_dir,save_type,elph_path,yambo_exec_path,converge_DG)
        
        #End IO        
        self.yf.IO_close()
        
        # Check if python must exit immediately or after all submitted jobs have completed
        if self.wait_up:
            if wait_all: wait_for_all_jobs(self.job_shells,self.job_folders)
            for shell in self.job_shells: shell.clean()
    
    def driver(self,STEPS,RUN,nscf_out,y_out_dir,save_type,elph_path,yambo_exec_path,converge_DG):
        """
        Worflow driver.
        
        It runs the following:
            - setup functions
            - job submission functions
            - double grid convergence functions
        """
        if STEPS=='1' or STEPS=='all': 
            self.yf.msg("------  STEP 1 ------")
            self.setup_cg()
            if RUN: self.run_jobs(nscf_out,y_out_dir)
        
        if STEPS=='2' or STEPS=='all':
            self.yf.msg("------  STEP 2 ------")
            for ig,cg in enumerate(self.cg_strings):
                # NSCF status check
                calc_dir  = '%s/%s_coarse_grid'%(self.nscf_dir,cg)            
                calc = self.check_nscf_completed(calc_dir,nscf_out)
                if calc: self.yf.msg("     NSCF CG %s found."%cg)
                else: self.yf.msg("     NSCF CG %s NOT found."%cg)
                # YAMBO status check
                ycalc_dir = '%s/%s_coarse_grid'%(self.yambo_dir,cg)
                ycalc = self.yambo_output_is_NOT_there(ycalc_dir,y_out_dir)
                if ycalc: self.yf.msg("     YAMBO CG %s NOT found."%cg)
                else: self.yf.msg("     YAMBO CG %s found."%cg)      
                #          
                if calc and ycalc:
                    yambo_dir = '%s/%s_coarse_grid'%(self.yambo_dir,cg)
                    CreateYamboSave(self.prefix,save_type=save_type,nscf=calc_dir,elph_path=elph_path,\
                                    database=os.path.abspath(yambo_dir),yambo_exec_path=yambo_exec_path,printIO=False)
                    self.setup_fg(calc_dir,yambo_dir,self.fg_grids[ig],self.fg_strings[ig])
            if RUN: self.run_jobs(nscf_out,y_out_dir)
                    
        if STEPS=='3' or STEPS=='all':
            self.yf.msg("------  STEP 3 ------")
            for ig,cg in enumerate(self.cg_strings):
                for iff,fg in enumerate(self.fg_strings[ig]):
                    # NSCF status check
                    calc_dir  = '%s/%s_coarse_grid/%s'%(self.nscf_dir,cg,fg)
                    calc = self.check_nscf_completed(calc_dir,nscf_out)
                    if calc: self.yf.msg("     NSCF CG %s FG %s found."%(cg,fg))
                    else: self.yf.msg("     NSCF CG %s FG %s NOT found."%(cg,fg))
                    # YAMBO status check
                    ycalc_dir = '%s/%s_coarse_grid/%s'%(self.yambo_dir,cg,fg)
                    ycalc = self.yambo_output_is_NOT_there(ycalc_dir,y_out_dir)
                    if ycalc: self.yf.msg("     YAMBO CG %s FG %s NOT found."%(cg,fg))
                    else: self.yf.msg("     YAMBO CG %s FG %s found."%(cg,fg))   
                    #      
                    if calc and ycalc:
                        yambo_dir = '%s/%s_coarse_grid/%s'%(self.yambo_dir,cg,fg)
                        if not os.path.isfile('%s/SAVE/ndb.Double_Grid'%yambo_dir):
                            CreateYamboSave(self.prefix,save_type='simple',nscf=calc_dir,elph_path=None,\
                                            database="%s/dg_SAVE"%os.path.abspath(yambo_dir),yambo_exec_path=yambo_exec_path,printIO=False)
                            self.setup_yambo_fg(yambo_dir,self.fg_grids[ig][iff],y_out_dir)
            if RUN: self.run_jobs(nscf_out,y_out_dir)
        
        # This is a plotting routine in the ip case. It has to be updated to a full convergence analysis and report.
        # The involved functions can possibly be moved into another file
        # 
        if (STEPS=='4' or STEPS=='all') and converge_DG :
            self.yf.msg("----------  STEP 4 ----------")
            self.yf.msg("-- Double grid convergence --")
            yout = 'o-%s.eps_q1_ip'%y_out_dir
            ip_data = []
            ip_labels = []
            #
            for ig,cg in enumerate(self.cg_strings):
                ip_data_temp = []
                ip_labels_temp = []
                temp_path = '%s/%s_coarse_grid/%s/%s'%(self.yambo_dir,cg,y_out_dir,yout)
                if os.path.isfile(temp_path):
                    self.yf.msg("     IP CG %s found."%cg)
                    ip_data_temp.append(np.loadtxt(temp_path))
                    ip_labels_temp.append(cg)
                else:
                    self.yf.msg("     IP CG %s NOT found."%cg)
                for iff,fg in enumerate(self.fg_strings[ig]):
                    temp_path = '%s/%s_coarse_grid/%s/%s/%s'%(self.yambo_dir,cg,fg,y_out_dir,yout)
                    if os.path.isfile(temp_path):
                        self.yf.msg("     IP CG %s FG %s found."%(cg,fg))
                        ip_data_temp.append(np.loadtxt(temp_path))
                        ip_labels_temp.append('%s N_FG=%d'%(cg,self.fg_grids[ig][iff]))
                    else:
                        self.yf.msg("     IP CG %s FG %s NOT found."%(cg,fg))
                ip_data.append(ip_data_temp)
                ip_labels.append(ip_labels_temp)
            #
            ip_data = [ip for ip in ip_data if ip != []]
            ip_labels = [ip for ip in ip_labels if ip != []]
            self.plot_ip_spectra(ip_data,ip_labels,self.E_laser)
        
    def setup_cg(self):
        """ First step of the workflow: setup CG folder tree and CG nscf calculations
        """
        for ig,cg in enumerate(self.cg_grids):
            work_dir = "%s/%s_coarse_grid"%(self.nscf_dir,self.cg_strings[ig])
            if not os.path.isdir(work_dir):
                os.mkdir(work_dir)
                self.generate_inputfile(work_dir,cg)
                os.system('cp -r %s/%s.save %s'%(self.scf_save_path,self.prefix,work_dir))
            yambo_dir = "%s/%s_coarse_grid"%(self.yambo_dir,self.cg_strings[ig])
            if not os.path.isdir(yambo_dir):
                os.mkdir(yambo_dir)
                self.ya_inp.write('%s/%s.in'%(yambo_dir,self.yambo_calc_type))
                
    def setup_fg(self,nscf_cg_dir,yambo_cg_dir,fg_grid,fg_string):
        """ Second step of the workflow: setup FG folder tree and FG (CG) nscf (yambo) calculations
        """
        for iff,fg in enumerate(fg_grid):
            fg_nscf_inp = '%s_fg.nscf'%self.prefix
            rand_nm = 'random_kpt_list_%s.dat'%fg_string[iff]
            ypp_inp = 'ypp_fg.in'
            work_dir = "%s/%s"%(nscf_cg_dir,fg_string[iff])
            if not os.path.isdir(work_dir):
                os.mkdir(work_dir)
                os.system('cp -r %s/%s.save %s'%(self.scf_save_path,self.prefix,work_dir))
            yambo_dir = "%s/%s"%(yambo_cg_dir,fg_string[iff])
            if not os.path.isdir(yambo_dir):
                os.mkdir(yambo_dir)
                os.mkdir("%s/dg_SAVE"%yambo_dir)
                self.ya_inp.write('%s/%s.in'%(yambo_dir,self.yambo_calc_type))
            if not os.path.isfile('%s/%s'%(work_dir,fg_nscf_inp)):
                self.generate_ypp_input_random_grid(yambo_cg_dir,fg,ypp_inp)
                ypp_run = self.frontend
                ypp_run.add_command('cd %s; %s -F %s > ypp_fg.log'%(yambo_cg_dir,self.ypp,ypp_inp))
                ypp_run.add_command('mv o.random_k_pts %s'%rand_nm)
                ypp_run.add_command('cp %s %s'%(rand_nm,work_dir))
                ypp_run.run()
                kpts_rndm = np.loadtxt('%s/%s'%(work_dir,rand_nm))
                if len(kpts_rndm) != fg:
                    self.yf.msg("[WARNING] Actual fine grid number of kpts is %d instead of %d"%(len(kpts_rndm),fg))
                self.generate_inputfile(work_dir,fg,klist=kpts_rndm)

    def setup_yambo_fg(self,yambo_fg_dir,fg_num,yresults_dir,clean_dg_saves=True):
        """ Third step of the workflow: map FG to CG and FG yambo calculations
        """
        ypp_inp = 'ypp_map.in'
        os_run = self.frontend
        if os.path.isfile('%s/../%s/ndb.dipoles'%(yambo_fg_dir,yresults_dir)):
            os_run.add_command('cd %s; cp ../%s/ndb.dipoles* ../SAVE/ ; cp -r ../SAVE .'%(yambo_fg_dir,yresults_dir))
        else:
            os_run.add_command('cd %s; cp -r ../SAVE .'%yambo_fg_dir)
        os_run.run()
        self.generate_ypp_input_map_grid(yambo_fg_dir,fg_num,ypp_inp)
        ypp_run = self.frontend
        ypp_run.add_command('cd %s; %s -F %s > ypp_map.log'%(yambo_fg_dir,self.ypp,ypp_inp))
        ypp_run.run()
        if os.path.isfile('%s/SAVE/ndb.Double_Grid'%yambo_fg_dir):
            self.yf.msg("  -> Double Grid ready. <-")
            if clean_dg_saves: 
                os.system('rm -r %s/dg_SAVE/SAVE'%yambo_fg_dir)
                os.system('touch %s/dg_SAVE/DOUBLE_GRID_SAVE_REMOVED_TO_SAVE_DISK_SPACE'%yambo_fg_dir)
        else:
            self.yf.msg("  -> Double Grid NOT ready. <-")
    
    def generate_inputfile(self,folder,kpts,klist=None):
        """ Modify nscf input file in case of CG or FG kpoint grids
        """
        import copy
        qe_input = copy.deepcopy(self.nscf_inp)
        qe_input.control['pseudo_dir'] = "'%s'"%self.pseudo_path
        
        if klist is None:
            qe_input.kpoints = kpts
        else:
            qe_input.ktype = "crystal"
            qe_input.klist = klist
            
        qe_input.write('%s/%s.nscf'%(folder,self.prefix))
    
    def generate_ypp_input_random_grid(self,folder,fg_num,inp_nm):
        """ Create ypp input file for the generation of the FG coordinates [units: yambo "rlu" -> qe "crystal"]
        """
        yppin = YamboIn()
        yppin.arguments.append('bzgrids')
        yppin.arguments.append('Random_Grid')
        yppin['OutputAlat'] = self.nscf_inp.system['celldm(1)']
        yppin['cooOut'] = "rlu"
        yppin['BZ_random_Nk'] = fg_num

        yppin.write('%s/%s'%(folder,inp_nm))
        
    def generate_ypp_input_map_grid(self,folder,fg_num,inp_nm):
        """ Create ypp input file for the mapping of the FG grid to the CG grid
        """
        yppin = YamboIn()
        yppin.arguments.append('kpts_map')
        yppin['FineGd_mode']='unexpanded'
        yppin['BZ_DbGd_Nk']=fg_num
        yppin.arguments.append('SkipCheck')
        yppin['FineGd_DB1_paths'] = ['./dg_SAVE']

        yppin.write('%s/%s'%(folder,inp_nm))
            
    def run_jobs(self,out_qe,out_yambo):
        """ Workflow
        """
        #MODULARIZATION ISSUE: 
        #       remember that function check_nscf_completed has a dependency 
        #       on the name of the qe output file - it has to be '*.out' - and 
        #       hence it depends on how this name is given in function shell_run     
                
        for ig,cg in enumerate(self.cg_strings): # ---------- Outer COARSE GRID loop ----------
            # temp_dir: where qe calculations are run
            # save_dir: where yambo calculations are run   

            temp_dir = '%s/%s_coarse_grid'%(self.nscf_dir,cg)
            
            if not self.check_nscf_completed(temp_dir,out_qe):
                self.yf.msg("Running NSCF CG %s..."%cg)      ############## Run NSCF CG
                self.qe_id_cg[ig] = self.shell_run("qe_%s"%cg,temp_dir,out_qe,'qe')
                if self.wait_up: self.job_folders.append(temp_dir)

            save_dir = '%s/%s_coarse_grid'%(self.yambo_dir,cg)
            if os.path.isdir('%s/SAVE'%save_dir):

                if self.yambo_output_is_NOT_there(save_dir,out_yambo) and not self.noyambo:
                    self.yf.msg("Running YAMBO CG %s..."%cg)     ############ Run YAMBO CG
                    self.ya_id_cg[ig] = self.shell_run("ya_%s"%cg,save_dir,out_yambo,'y')    # depends on JOBID='%d'%self.qe_id_cg[ig])
                    if self.wait_up: self.job_folders.append(save_dir)

            try: self.fg_strings[ig]
            except IndexError as err: raise Exception('No fine grid(s) provided for CG %s.'%cg) from err 
                      
            for iff,fg in enumerate(self.fg_strings[ig]):      # ---------- Inner FINE GRID loop ----------

                temp_dir = '%s/%s_coarse_grid/%s'%(self.nscf_dir,cg,fg)
                if os.path.isdir(temp_dir):

                    if not self.check_nscf_completed(temp_dir,out_qe):
                        self.yf.msg("Running NSCF CG %s FG %s..."%(cg,fg)) ######### Run NSCF FG
                        self.qe_id_fg[ig][iff] = self.shell_run("qe_%s"%cg,temp_dir,out_qe,'qe') # depends on JOBID='%d'%self.qe_id_cg[ig])
                        if self.wait_up: self.job_folders.append(temp_dir)

                save_dir = '%s/%s_coarse_grid/%s'%(self.yambo_dir,cg,fg)
                if os.path.isdir('%s/SAVE'%save_dir) and os.path.isfile('%s/SAVE/ndb.Double_Grid'%save_dir):
                    
                    if self.yambo_output_is_NOT_there(save_dir,out_yambo) and not self.noyambo:
                        self.yf.msg("Running YAMBO CG %s FG %s..."%(cg,fg))     ############ Run YAMBO FG
                        self.ya_id_fg[ig][iff] = self.shell_run("ya_%s"%cg,save_dir,out_yambo,'y') # depends on JOBID='%d:%d'%(self.ya_id_cg[ig],self.qe_id_fg[ig][iff]))
                        if self.wait_up: self.job_folders.append(save_dir)

    def shell_run(self,jname,run_dir,out_dir,exec,JOBID=None):
        """ 
        Submit job
        
            exec:
                qe: runs pw.x
                y:  runs the yambo* executable of choice
            
            jname: job name
            JOBID: job id of simulation that the present job has a dependency on
            run_dir: where job is run
            out_dir: name of output (folder and log for yambo; '%s.out'%out_dir file for qe)  
            
            returns id of present submitted job          
        """
        if exec=='qe': shell = deepcopy(self.qejobrun)
        if exec=='y':  shell = deepcopy(self.yjobrun)
        shell.name = '%s_%s'%(jname,shell.name)
        
        # Add dependency if specified
        if self.wait_up and JOBID is not None:
            dependency='afterok:%s'%JOBID
            shell.dependency=dependency
        
        if exec=='qe': shell.add_mpirun_command('%s -inp %s.nscf > %s.out'%(self.pw,self.prefix,out_dir))
        if exec=='y': shell.add_mpirun_command('%s -F %s.in -J %s -C %s 2> %s.log'%(self.yambo,self.yambo_calc_type,out_dir,out_dir,out_dir))
        shell.run(filename='%s/%s.sh'%(run_dir,exec)) ### Specify run path
        
        # Manage submissions if specified
        #if self.wait_up: wait_for_job(shell,run_dir)
        if self.wait_up: this_job_id = shell.jobid
        else:            this_job_id = -1
        if self.wait_up: self.job_shells.append(shell) 
        else: shell.clean()
        
        return this_job_id
    
    def ip_input_is_there(self):
        """ Check if yambo ip input is correctly given in converge_DG case
        """
        condition = 'chi' in self.ya_inp.arguments and \
                    'optics' in self.ya_inp.arguments and \
                    self.ya_inp['Chimod']=='IP'
        if not condition:
            raise FileNotFoundError("IP input file not recognised: are you sure you specified 'chi' and 'optics' runlevels and Chimod='IP'?")

    def check_nscf_completed(self,folder,out_prefix):
        """ Check if nscf calculation has completed and proceed
        """
        status = True
        try:
            check = subprocess.check_output("grep JOB %s/%s*"%(folder,out_prefix), shell=True, stderr=subprocess.STDOUT)
            check = check.decode('utf-8')
            check = check.strip().split()[-1]
        except subprocess.CalledProcessError as e:
            check = ""
        if check != "DONE.": status = False
        return status

    def yambo_output_is_NOT_there(self,calc_dir,results_dir):
        """ Check if yambo produced non-empty outputs
        """
        condition = 0
        out_files = glob('%s/%s/o-%s.*'%(calc_dir,results_dir,results_dir))
        for out_file in out_files:
            test = ( not os.path.isfile(out_file) ) or ( os.path.isfile(out_file) and os.stat(out_file).st_size == 0 )
            if test: condition+=1
        if condition==len(out_files): return True # This means no output file has been produced: this calculation must be run
        elif condition==0: return False # This means all output files have been produced: this calculation must be skipped
        else: raise UserWarning('Some output files have been produced and some not -- check the above calculation.') 
    
    #TODO: move convergence-related functions to a new submodule
    def plot_ip_spectra(self,data,labels,w_laser,fig_name='ip_spectra',xlims=None):
        """ Plot results in a dynamic layout (chosen by FP)
        """
        if not os.path.isdir(self.plot_dir): os.mkdir(self.plot_dir)
        lwidth = 0.8
        f, (axes) = plt.subplots(1,len(data),sharey=True,figsize=(12,4))

        for ig,cg in enumerate(data):
            colors = plt.cm.gist_rainbow(np.linspace(0.,1.,num=len(cg)))
            axes[ig].set_ylim(0.,1.1*np.max(cg[0][:,1]))
            if xlims is not None: axes[ig].set_xlim(xlims[0],xlims[1])
            for iff,fg in enumerate(cg):
                axes[ig].plot(fg[:,0],fg[:,1], '-', lw = lwidth, color = colors[iff], label = labels[ig][iff])
            axes[ig].fill_between(cg[0][:,0], cg[-1][:,1], color = colors[-1], alpha = 0.2)

        for ax in axes:
            ax.axvline(w_laser,lw=0.8,color='black',zorder=-1)
            ax.legend()

        f.tight_layout()

        plt.savefig('%s/%s.png'%(self.plot_dir,fig_name),format='png',dpi=150)
    
    def clean_everything(self):
        """ Remove workflow tree
        """
        rm_run = self.frontend
        rm_run.add_command('rm -rf %s %s %s'%(self.nscf_dir,self.yambo_dir,self.plot_dir))
        rm_run.run()
        self.yf.msg("Workflow removed.")
    
    def branch_wise_flow(self):
        """
        The workflow dependencies are complicated.
        We understand them in two steps.
        
        First, how the actual job submissions depend on each other:
        
                qe_CG_1                     qe_CG_2              ...            qe_CG_n
                   |                          |                                    |
             ______|________________          |                                    |    
             |         |           |       (same in parallel)                   (same in parallel)
             |         |           |       
        qe_FG_1  ... qe_FG_n    y_CG_1
           |           |______ ... | ... _____
           |_______________________|_________|
                              |              |
                              |              |
                            y_FG_1   ...   y_FG_n      
                            
                            if converge_DG: REDUX ALL JOBS
                                                  |
                                                  |
                                                PLOTS  
                            
        All the separate branches of the workflow could be run sequentially using the dependency system of the scheduler.
        
        Second, how the workflow functions actually depend on each other (single tree branch is shown here):
        
                        qe_CG_1           
                           | 
                     ______|________________                    first barrier   
                     |         |           | 
                     |         |           |
                     |---------|--------- SAVE   
                     |         |           |                    second barrier
                setup_fg    setup_fg       |                (very small delay here)
                     |         |           |                    third barrier
                     |         |           |
                qe_FG_1  ... qe_FG_n    y_CG_1
                   |           |______ ... | ... _____
                   |_______________________|_________|          fourth barrier
                                      |              |
                                      |              | 
                                setup_yambo_fg  setup_yambo_fg
                                      |              |          fifth barrier
                                      |              |
                                    y_FG_1   ...   y_FG_n 
                                    
        For this reason, it's better to plan a submission 'STEP by STEP' of the workflow instead than 'branch by branch'. 
        However, the latter way would be more efficient if implemented. 
        """



# This script reads the carrier database
# and display it along a path in histogram form
# along with a representation of the carriers in energy space

from __future__ import print_function
from __future__ import division
from builtins import range
from past.utils import old_div
from yambopy import *
import matplotlib.gridspec as gridspec
from scipy.optimize import curve_fit
import os

############
# SETTINGS #
############

folder = 'rt-24x24'
calc   = 'QSSIN-D-100.0fs-2.07eV-300K-DG' # Where RT carrier output is
path = [[0.0,0.0,0.0],[0.5,0.0,0.0],[0.33333,0.33333,0.0],[0.0,0.0,0.0]]
nbv = 2 ; nbc = 2 # nb of valence and conduction bands

########
# INIT #
########

# For saving pictures
os.system('mkdir -p occupations_v2/%s/%s'%(folder,calc))

# Instance containing bandstructure (as used in RT sim) and occupations
yrt = YamboRTDB(folder=folder,calc=calc)

yrt.get_path(path) # Generates kindex and distances

### aliases
times = [i * 1e15 for i in yrt.times] # carriers output times, in fs
nbands = yrt.nbands # number of bands in the RT simulation

if nbv+nbc != nbands:
    raise NameError('Incompatible number of bands, set nbv and nbc in script.')

## 'path-plot' variables
kindex = yrt.bands_indexes # kpoint indexes (in order) to draw path
distances = yrt.bands_distances
eigenvalues = yrt.eigenvalues[kindex,:] # eigenvalues of the bands included in the RT simulation
#
max_occ = np.amax(yrt.occupations[:,kindex,:]) # used to size the distribution plots
occupations = old_div(yrt.occupations[:,kindex,:],max_occ) # format time,kindex,band index (from 0 to nbands, only on path)
#
##

## 'fit' variables and function
# FD distrib for fit
def fermi_dirac(E,a,T): # declare E first for fit
    return old_div(1,(1+np.exp(old_div((E-a),T))))
#
KtoeV = 8.61733e-5
#
# xeng is an array of values to plot the fit properly
xeng = np.linspace(np.amin(eigenvalues[:,list(range(nbv))]), np.amax(eigenvalues[:,list(range(nbv,nbands))]),1000)
##

##############
# EXT. FIELD #
##############

# The external field is read from the o- file
ext = np.loadtxt('%s/%s/pulse/o-pulse.external_field'%(folder,calc))
field = old_div(ext[:,2],max(abs(ext[:,2]))) # polarization : x=1,y=2,z=3

##################
# ENERGY DISTRIB #
##################

# Sort the (n,k) pairs between positive and negative energies
# (If the same energy appears twice, it must not be summed over)
list_e=[] ; list_h=[]
for k in range(yrt.nkpoints):
    for n in range(yrt.nbands):
        e = yrt.eigenvalues[k,n]
        if e<=0.0:
            list_h.append((k,n))
        else:
            list_e.append((k,n))

# Map (k,n) to a single index for e and h
# then get the list of indices to sort the energies
nrj = np.zeros((len(list_e)))
for i,(k,n) in enumerate(list_e):
    nrj[i]=yrt.eigenvalues[k,n]
sorted_e = np.argsort(nrj)

nrj = np.zeros((len(list_h)))
for i,(k,n) in enumerate(list_h):
    nrj[i]=yrt.eigenvalues[k,n]
sorted_h = np.argsort(nrj)


# Build the occupation tables occ_x[t,(nk)_index,(e|occ)]
occ_e = np.zeros((len(times),len(list_e),2))
for t in range(len(times)):
    for i,(k,n) in enumerate(list_e):
        occ_e[t,i,0]=yrt.eigenvalues[k,n]
        occ_e[t,i,1]=yrt.occupations[t,k,n]

occ_h = np.zeros((len(times),len(list_h),2))
for t in range(len(times)):
    for i,(k,n) in enumerate(list_h):
        occ_h[t,i,0]=yrt.eigenvalues[k,n]
        occ_h[t,i,1]=yrt.occupations[t,k,n]

# Sorting
for t in range(len(times)):
    occ_e[t,:,:]=occ_e[t,sorted_e,:]
    occ_h[t,:,:]=occ_h[t,sorted_h,:]

# *(-1) on holes to fit the same way as electrons
occ_h *= -1



#################
# BAR PLOT DATA #
#################

# Fill arrays with occupation of valence and conduction bands
# Recall that 'occupations' was normalized
# one entry per band
occ_v = np.zeros((len(times),len(kindex),nbv))
occ_c = np.zeros((len(times),len(kindex),nbc))
for n in range(nbv):
    occ_v[:,:,n] = -occupations[:,:,n] # minus sign to get positive occupations
for n in range(nbc):
    occ_c[:,:,n] = occupations[:,:,n+nbv] # +nbv to read CBs



####################
# TIME LOOP & PLOT #
####################

# Gridspec allows to place subplots on a grid
# spacing for exemple can be customised
gs = gridspec.GridSpec(9, 8)

# y range for band structure & energy plots
ymin_v= np.amin(eigenvalues[:,:nbv])-0.1
ymin_c= np.amin(eigenvalues[:,nbv:])-0.1

ymax_v= np.amax(eigenvalues[:,:nbv])+0.1
ymax_c= np.amax(eigenvalues[:,nbv:])+0.1
###

for t in range(len(times)):
#for t in (30,):
    i=t
    print(times[i])

    name = 'occupations_v2/'+folder+'/'+calc+'/%d.png' % (times[t])

    fig = plt.figure()

    fig.suptitle('Occupation of the bands and fit to the Fermi-Dirac distribution',fontsize=14,ha='center')

####### bandstructure w/ occupation plot

    ax1c = plt.subplot(gs[0:4,0:-2])
    ax1v = plt.subplot(gs[4:8,0:-2])
    # remove x ticks
    ax1c.tick_params(axis='x',which='both',bottom='off',top='off',labelbottom='off')
    ax1v.tick_params(axis='x',which='both',bottom='off',top='off',labelbottom='off')

    # set x range
    ax1c.set_xlim((0,distances[-1]))
    ax1v.set_xlim((0,distances[-1]))
    # y range is defined with ax3 and ax4 (they share y axis with ax1)


    # Plot band structure
    ax1v.plot(distances,eigenvalues[:,:nbv],'k-',lw=2,zorder=0)
    ax1c.plot(distances,eigenvalues[:,nbv:],'k-',lw=2,zorder=0)

    # VB
    for n in range(nbv):
        ax1v.scatter(distances,eigenvalues[:,n],s=400*occ_v[t,:,n],color='blue',alpha=0.5)
    # CB
    for n in range(nbc):
        ax1c.scatter(distances,eigenvalues[:,nbv+n],s=400*occ_c[t,:,n],color='red',alpha=0.5)

    # text and labels
    fig.text(0.05,0.6,'Energy (eV)',size=16,rotation='vertical')
    fig.text(0.50,0.91, '%d fs'%times[t],size=16)

######## field plot

    ax2 = plt.subplot(gs[-1,:])

    # remove ticks and labels
    ax2.tick_params(axis='x',which='both',bottom='off',top='off',labelbottom='off')
    ax2.tick_params(axis='y',which='both',left='off',right='off',labelleft='off')

    # text
    ax2.set_ylabel('Field')

    # frame size
    ax2.set_xlim((0,times[-1]))
    ax2.set_ylim((-1.3,1.3))

    ax2.plot(field[:int(times[t])])

## Plot of the occupation as a function of energy (rotated to match the band structure)
    ax3 = plt.subplot(gs[0:4,-2:],sharey=ax1c)
    ax4 = plt.subplot(gs[4:8,-2:],sharey=ax1v)

# plot the data
    try: # does not break if fit is not found
        fit,cov = curve_fit(fermi_dirac,occ_e[i,:,0],occ_e[i,:,1])
    except RuntimeError:
        fit=np.array([0,0])

    ax3.fill_betweenx(occ_e[i,:,0],0,occ_e[i,:,1],color='red')
    ax3.plot(fermi_dirac(xeng,fit[0],fit[1]),xeng,'k-')
    ax3.text(0.5,0.9,'Electrons\nT = %d K'%(old_div(fit[1],KtoeV)),transform=ax3.transAxes,ha='center',va='center')

    try:
        fit,cov = curve_fit(fermi_dirac,occ_h[i,:,0],occ_h[i,:,1])
    except RuntimeError:
        fit=np.array([0,0])

    ax4.fill_betweenx(-occ_h[i,:,0],0,occ_h[i,:,1],color='blue')
    ax4.plot(fermi_dirac(xeng,fit[0],fit[1]),-xeng,'k-')
    ax4.text(0.5,0.1,'Holes\nT = %d K'%(old_div(fit[1],KtoeV)),transform=ax4.transAxes,ha='center',va='center')

# set x and y range
    ax4.set_xlim(-0.1*max_occ,1.1*max_occ)
    ax3.set_xlim(-0.1*max_occ,1.1*max_occ)

    ax3.set_ylim(( ymin_c,ymax_c ))
    ax4.set_ylim(( ymin_v,ymax_v ))

# hide some ticks/labels
    ax3.tick_params(axis='x',which='both',bottom='off',top='off',labelbottom='off')
    ax3.tick_params(axis='y',labelleft='off',labelright='off')

    ax4.tick_params(axis='x',which='both',bottom='off',top='off',labelbottom='off')
    ax4.tick_params(axis='y',labelleft='off',labelright='off')

    plt.savefig( name ,transparent=False,dpi=300)
    print(name)
    #plt.show()
    plt.close(fig)



from yambopy import *
from yambopy.plot  import *

# Read the yambo_nl databases for all laser frequencies
NLDB=YamboNLDB()

#
# Analize the non-linear reponse using Eq.
# of Eq. 26 and 27 of Ref. Phys. Rev. B 88, 235113 (2013)
# results are stored in the files o.YamboPy-X_probe_order_x 
# where x goes from 0 to X_order
#
Harmonic_Analysis(NLDB,X_order=4)



# Copyright (C) 2018 Henrique Pereira Coutada Miranda
# All rights reserved.
#
# This file is part of yambopy
#
import unittest
import os
from qepy.pw import PwIn
from yambopy.data.structures import Si

class TestPwIn(unittest.TestCase):

    def test_pwin(self):
        pwi = PwIn.from_structure_dict(Si) 
        pwi.set_nscf(10)
        print(pwi)

        pwi.cell_parameters = [[1,0,0],[0,1,0],[0,0,1]]
        pwi.ibrav = 0
        print(pwi)

if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2023, Claudio Attaccalite
# All rights reserved.
#
# This file is part of the yambopy project
# Calculate linear response from real-time calculations (yambo_nl)
#
import numpy as np
from yambopy.nl.fft_interp import *
from yambopy.nl.external_efield import *
from yambopy.units import ha2ev,fs2aut

from matplotlib.ticker import FuncFormatter
import matplotlib.pyplot as plt
import os

def sci_format(x,lim):
    return '{:.1e}'.format(x)

def Plot_Pol_or_Curr(time=None, pol=None, curr=None, xlim=None,save_file=None):
    if not isinstance(pol, np.ndarray) and isinstance(pol, np.ndarray):
        print("Polarzation or Current not present")
        return
    if not isinstance(time, np.ndarray):
        print("Time series not present")
        return

    char_size=14


    fig=plt.figure(figsize=(10,10))
    fig.subplots_adjust(hspace=.05)

    major_formatter = FuncFormatter(sci_format)

    if xlim == None:
        xlim=[0.0,time[-1]/fs2aut]

    if isinstance(pol, np.ndarray):
        pj='P'
        arr=pol
        fig.suptitle(' Real-time polarization in the three cartesian directions ', fontsize=char_size)
    else:
        pj='J'
        arr=curr
        fig.suptitle(' Real-time current in the three cartesian directions ', fontsize=char_size)

    ax=plt.subplot(3,1,1)
    plt.tick_params(axis='both', which='major', labelsize=char_size)
    plt.plot(time[:]/fs2aut,arr[0,:],color="blue", linewidth=1.5, linestyle="-",label=pj+"$_x$(t)")
    ax.yaxis.set_major_formatter(major_formatter)
    ax.set_xticks([])
    ax.set_xlim(xlim)
    ax.legend(fontsize=char_size)

    ###########6
    ax=plt.subplot(3,1,2)
    plt.tick_params(axis='both', which='major', labelsize=char_size)
    plt.plot(time[:]/fs2aut,arr[1,:],color="blue", linewidth=1.5, linestyle="-",label=pj+"$_y$(t)")
    ax.yaxis.set_major_formatter(major_formatter)
    ax.set_xticks([])
    ax.set_xlim(xlim)
    ax.legend(fontsize=char_size)
    
    ax=plt.subplot(3,1,3)
    plt.tick_params(axis='both', which='major', labelsize=char_size)
    plt.plot(time[:]/fs2aut,arr[2,:],color="blue", linewidth=1.5, linestyle="-",label=pj+"$_z$(t)")
    ax.set_xlabel('[fs]',size=14)
    ax.set_xlim(xlim)
    ax.legend(fontsize=char_size)
    if save_file is not None:
        plt.savefig(save_file)

    plt.show()

def Linear_Response(time, pol, efield, pol_ref=None, e_range=[0.0, 20.0], n_freqs=200, 
        output_file="o.YamboPy-eps_along_E",plot=False,plot_file=None):
    if efield["name"] != "DELTA":
        print("Linear response implemented only for Delta function external fields ")
        sys.exit(0)

    # Pump and probe minus the pump if present
    if pol_ref is not None:
        pol=pol-pol_ref

    freqs=np.linspace(e_range[0],e_range[1],n_freqs)/ha2ev
    pol_w=np.zeros((pol.shape[0],n_freqs),dtype=complex)

    efield_w=get_Efield_w(freqs,efield)

    Fourier_Interpolation(pol,pol_w,time,freqs,mode="T2W")
    #
    # Polarization along the field direction
    #
    pol_w_along_E=np.zeros(n_freqs,dtype=complex)
    for i_d in range(3):
        pol_w_along_E[:]+=pol_w[i_d,:]*efield["versor"][i_d]
    #
    # EPS = 1+4.0*pi*Xhi(omega)
    #
    eps=np.zeros(n_freqs,dtype=complex)
    eps=1.0+4.0*np.pi*pol_w_along_E/efield_w

    header="E [eV]      Im/eps      Re/eps"
    footer='Linear response analysis performed using YamboPy'
    np.savetxt(output_file,np.c_[freqs*ha2ev,eps.imag,eps.real],header=header,delimiter=' ',footer=footer)

    fig, axs = plt.subplots(2)
    fig.suptitle('Dielectric constant along the field direction')
    axs[0].set_xlim(e_range)
    axs[0].plot(freqs*ha2ev,eps.real,label='Real part ')
    axs[0].legend()
    axs[0].set_xticks([])
    axs[1].set_xlim(e_range)
    axs[1].plot(freqs*ha2ev,eps.imag,label='Imag part ')
    axs[1].legend()
    axs[1].set_xlabel('eV')
    if plot:
        plt.show()
    if plot_file is not None:
        plt.savefig(plot_file)



#
# License-Identifier: GPL
#
# Copyright (C) 2024 The Yambo Team
#
# Authors: HPC
#
# This file is part of the yambopy project
#
import numpy as np

def calculate_distances(kpoints):
    """
    take a list of k-points and calculate the distances between all of them
    """
    kpoints = np.array(kpoints)
    distances = [0]
    distance = 0
    for nk in range(1,len(kpoints)):
        distance += np.linalg.norm(kpoints[nk-1]-kpoints[nk])
        distances.append(distance)   
    return distances

class Path(object):
    """ Class that defines a path in the brillouin zone
       
        Input format:
        :: Path(klist,intervals)

        :: klist = [ [[ k1_x, k1_y, k1_z ], 'K1_label'],
                        ...                 ...
                     [[ kN_x, kN_y, kN_z ], 'KN_label'] ]
        
        :: intervals = [ n_steps_1, n_steps_2, ... , n_steps_N-1 ]

        NB: interval steps are the number of kpoints along each high-symmetry direction, must be provided but is USED ONLY FOR INTERPOLATIONS
    """
    def __init__(self,klist,intervals):
        """
        Generation of a path in reciprocal space by specifying a list of k-points
        """

        self.intervals = intervals
        if len(intervals)>len(klist)-1:
            print(f"[WARNING] number of intervals is {len(intervals)}, should be {len(klist)-1} (no. of path lines). Taking the first {len(klist)-1} intervals.")
            self.intervals = intervals[:len(klist)-1]
        if len(intervals)<len(klist)-1:
            Nsteps=20
            print(f"[WARNING] number of intervals (path lines) inconsistent with special points specified, using default of {Nsteps} steps per direction")
            self.intervals = [ Nsteps for ik in range(len(klist)-1) ]

        klabels = []
        kpoints = []
        for kpoint, klabel in klist:
            kpoints.append(kpoint)
            klabels.append(klabel)
        self.kpoints = np.array(kpoints)
        self.klabels = klabels
    
    def as_dict(self):
        d = {'kpoints':self.kpoints.tolist(),
             'klabels':self.klabels,
             'intervals':self.intervals}
        return d

    @classmethod
    def from_dict(cls,d):
        klist = zip(d['kpoints'],d['klabels'])
        return cls(klist,d['intervals'])

    @property
    def distances(self):
        if hasattr(self,'_distances'): return self._distances
        k0 = np.array(self.kpoints[0])
        dist = 0
        distances = [0]
        for kpt in np.array(self.kpoints[1:]):
            dist += np.linalg.norm(kpt-k0)
            k0 = kpt
            distances.append(dist)
        self._distances = distances
        return distances

    def set_xticks(self,ax):
        ax.set_xticks(self.distances)
        ax.set_xticklabels(self.klabels)

    def __iter__(self):
        return iter(zip(self.kpoints,self.klabels,self.distances))

    def get_klist(self):
        """ 
        Output in the format of quantum espresso == [ [kx, ky, kz, 1], ... ]
        """
        kpoints = self.kpoints
        intervals = self.intervals
        kout  = np.zeros([sum(intervals)+1,4])
        kout[:,3] = 1
        io = 0
        for ik,interval in enumerate(intervals):
          for ip in range(interval):
            kout[io,:3] = kpoints[ik] + float(ip)/interval*(kpoints[ik+1] - kpoints[ik])
            io = io + 1
        kout[io,:3] = kpoints[ik] + float(ip+1)/interval*(kpoints[ik+1] - kpoints[ik])

        return kout

    def get_indexes(self):
        """ get the index of each point of the path
        """

        indexes = []
        index = 0
        for n,label in enumerate(self.intervals):
            indexes.append([index,self.klabels[n]])
            index += self.intervals[n] 
        indexes.append([index,self.klabels[-1]])
        return indexes

def vec_in_list(veca,vec_list,atol=1e-6):
    """ check if a vector exists in a list of vectors
    """
    return np.array([ np.allclose(veca,vecb,rtol=atol,atol=atol) for vecb in vec_list ]).any()

def red_car(red,lat):
    """
    Convert reduced coordinates to cartesian
    """
    lat = np.array(lat)
    red = np.array(red)
    if lat.shape != (3,3):
        raise ValueError('Wrong lattice dimensions expected (3,3) got {}'.format(lat.shape))
    return np.array([coord[0]*lat[0]+coord[1]*lat[1]+coord[2]*lat[2] for coord in red])

def car_red(car,lat):
    """
    Convert cartesian coordinates to reduced
    """
    car = np.array(car)
    lat = np.array(lat)
    if lat.shape != (3,3):
        raise ValueError('Wrong lattice dimensions expected (3,3) got {}'.format(lat.shape))
    return np.array([np.linalg.solve(np.array(lat).T,coord) for coord in car])

def rec_lat(lat):
    """
    Calculate the reciprocal lattice vectors
    """
    a1,a2,a3 = np.array(lat)
    v = np.dot(a1,np.cross(a2,a3))
    b1 = np.cross(a2,a3)/v
    b2 = np.cross(a3,a1)/v
    b3 = np.cross(a1,a2)/v
    return np.array([b1,b2,b3])



# Copyright (C) 2018 Henrique Pereira Coutada Miranda
# All rights reserved.
#
# This file is part of yambopy
#
#
"""
Schedulerpy is a simple set of python modules to run 
commands on different environments (clusters, local computers, etc..)

 **Currently available schedulers are:**

 * bash: Execute the job in the bash
 * oar : Use the OAR scheduler
 * pbs : Use the PBS scheduler

"""
from schedulerpy.scheduler import *
from schedulerpy.oar import *
from schedulerpy.pbs import *
from schedulerpy.bash import *



#
# License-Identifier: GPL
#
# Copyright (C) 2024 The Yambo Team
#
# Authors: HPC, AMS, FP, RR
#
# This file is part of the yambopy project
#
import re
import xml.etree.ElementTree as ET
from numpy import array, zeros, pi, conjugate, arange
from .lattice import Path, calculate_distances 
from yambopy.tools.string import marquee
from .auxiliary import *
from itertools import chain

RytoeV = 13.605698066
HatoeV = 2.0*RytoeV

class ProjwfcXML(object):
    """
    Class to read data from a Quantum espresso projwfc XML file.
    
    This file contains the projection of the Kohn-Sham stated in
    the atomic orbitals read from the pseudopotential
    """
    _proj_file = 'atomic_proj.xml'

    def __init__(self,prefix,output_filename='projwfc.log',path='.',qe_version='7.0'):
        """
        Initialize the structure with the path where the atomic_proj.xml is
        The zero energy is fixed at the Fermi energy
        Check conversion of Fermi energy (Ry to eV or Ha to eV)
        """
        self.qe_version   = qe_version
        self.prefix       = prefix
        self.path         = path
        self.get_qe_version(path,output_filename)
        version_number = float(re.findall(r"([0-9.]+)",self.qe_version)[0][:3])
        if version_number<6.7: raise NotImplementedError(f"Warning: QE version {self.qe_version} not supported by this script (needs to be>=6.7).")
        if version_number>=7.0: self.order_is_l_j_mj   = True
        else:                   self.order_is_j_l_m_mj = True
        self.datafile_xml = ET.parse( "%s/%s.save/%s"%(path, prefix, self._proj_file)).getroot()

        # Read the number of BANDS
        self.nbands = int( self.datafile_xml.findall("HEADER")[0].attrib['NUMBER_OF_BANDS'] )
        #get nkpoints
        self.nkpoints = int( self.datafile_xml.findall("HEADER")[0].attrib['NUMBER_OF_K-POINTS'] )
        #get number of spin components
        self.spin_components = int( self.datafile_xml.findall("HEADER")[0].attrib['NUMBER_OF_SPIN_COMPONENTS'] )
        #get number of projections
        self.nproj = int( self.datafile_xml.findall("HEADER")[0].attrib['NUMBER_OF_ATOMIC_WFC'] )
        #get fermi
        self.fermi = float( self.datafile_xml.findall("HEADER")[0].attrib['FERMI_ENERGY'] )*RytoeV
        #self.number_electrons = int( self.datafile_xml.findall("HEADER")[0].attrib['NUMBER_OF_ELECTRONS'])
           
        #get kpoints
        self.kpoints = self.get_kpoints()

        # Read Eigenvalues

        if self.spin_components == 1: self.eigen = self.get_eigen()
        if self.spin_components == 2: self.eigen1,self.eigen2  = self.get_eigen()
        if self.spin_components == 4: self.eigen = self.get_eigen()

        # Read Atomic Orbitals Projections
          
        if self.spin_components == 1: self.proj  = self.get_proj()
        if self.spin_components == 2: self.proj1,self.proj2 = self.get_proj()
        if self.spin_components == 4: self.proj  = self.get_proj()

        #here we open the ouput file of projwfc and get the quantum numbers of the orbitals
        try:
            # Try to open the file in the specified path
            f = open(os.path.join(path, output_filename), 'r')
        except FileNotFoundError:
            try:
                # Try to open the file in the './out' subdirectory of the specified path
                f = open(os.path.join(path, './out', output_filename), 'r')
            except FileNotFoundError:
                # Raise an exception if the file is not found in both locations
                raise Exception(f"The output file of projwfc.x: {output_filename} was not found in the directory or its ./out subdirectory.")        
            
        if hasattr(self, 'order_is_l_j_mj') and self.order_is_l_j_mj:
            states = []
            #                                                                                         wfc                  l                 j                 m_j                 
            for line in re.findall('state\s+\#\s+([0-9]+):\s+atom\s+([0-9]+)\s+\(([a-zA-Z0-9]+)\s*\),\s+wfc\s+([0-9])\s+\((?:l=([0-9.]+))? ?(?:j=([0-9.]+))? ?(?:m_j=\s+([0-9.]+))?',f.read()):
                # examples of the lines we have to read
                #  5: atom   1 (C  ), wfc  3 (l=2 m= 1)               #no spin case
                #  5: atom   1 (C  ), wfc  3 (j=1.5 l=1 m_j=-1.5)     #non collinear spin case
                istate, iatom, atype, wfc, l, j, m_j = line
                if j: j = float(j)
                if l: l = int(l)
                if m_j: m_j = float(m_j)
                states.append({'istate':int(istate), 'iatom':int(iatom), 'atype':atype, 'wfc':int(wfc), 'l':l, 'j':j, 'm_j':m_j})
            self.states = states

            f.close() 

        if hasattr(self, 'order_is_j_l_m_mj') and self.order_is_j_l_m_mj:      
            states = []
            #                                                                                        wfc                  j                 l                 m                    m_j
            for line in re.findall('state\s+\#\s+([0-9]+):\s+atom\s+([0-9]+)\s+\(([a-zA-Z0-9]+)\s*\),\s+wfc\s+([0-9])\s+\((?:j=([0-9.]+))? ?(?:l=([0-9.]+))? ?(?:m=\s+([0-9.]+))? ?(?:m_j=([ \-0-9.]+))?',f.read()):
                # examples of the lines we have to read
                #  5: atom   1 (C  ), wfc  3 (l=2 m= 1)               #no spin case
                #  5: atom   1 (C  ), wfc  3 (j=1.5 l=1 m_j=-1.5)     #non collinear spin case
                istate, iatom, atype, wfc, j, l, m, m_j = line
                if j: j = float(j)
                if l: l = int(l)
                if m: m = int(m)
                if m_j: m_j = float(m_j)
                states.append({'istate':int(istate),'iatom':int(iatom), 'atype':atype, 'wfc':int(wfc), 'j':j, 'l':l, 'm':m, 'm_j':m_j})
            self.states = states

            f.close()
    
    def get_qe_version(self,path,output_filename):
        """
        Get the version of Quantum Espresso. v<6.7 is not supported.

        Try both data-file-schema and projwfc output
        """
        try:
            datafile_schema_xml = ET.parse( f"{path}/{self.prefix}.save/data-file-schema.xml").getroot()
            qe_version_xml = datafile_schema_xml.findall("general_info/creator")[0].attrib["VERSION"].strip()
        except FileNotFoundError: qe_version_xml = None
        try:
            with open(f"{path}/{output_filename}") as f: qe_version_output = re.findall(r"Program PROJWFC v.([0-9.]+[a-zA-Z]*)", f.read())[0]
        except FileNotFoundError: qe_version_output = None
        if qe_version_xml and qe_version_output: 
            if qe_version_xml != qe_version_output: print(f"Warning: QE version mismatch between data-file-schema.xml ({qe_version_xml}) and projwfc output ({qe_version_output}). Using the output version for compatibility.")
            self.qe_version = qe_version_output
        elif qe_version_xml:    self.qe_version = qe_version_xml
        elif qe_version_output: self.qe_version = qe_version_output
        else: print(f"Warning: QE version not found in data-file-schema.xml or projwfc output. Using default version {self.qe_version}.")

    def get_indexes(self):
        """
        Get indexes of the bands where the projection is maximal
        """
        # Selection of the bands
        proj = zeros([self.nkpoints,self.nproj],dtype=int)
        for ik in range(self.nkpoints):
            for ip in range(self.nproj):
                proj[ik,ip] = np.argmax(np.absolute(self.proj[ik,ip,:])**2)

        return proj

    def get_states_helper(self, atom_query=['all'], orbital_query=['s','p','d','f']):
        """
        Get the sates that you want based on dictionary query by providing array of atoms and orbitals, default all orbitals
        
        Returns an array with the indices of the requested states in the qe array
        """
        states =  self.states
        queried_states = []

        for state in states:
            if (atom_query == ['all']) or (state['atype'] in atom_query):
                if (state['l']==0) and 's' in orbital_query:         #s orbital
                    queried_states.append(state['istate'] - 1)
                if (state['l']==1) and 'p' in orbital_query:         #p orbital
                    queried_states.append(state['istate'] - 1)
                if (state['l']==2) and 'd' in orbital_query:         #d orbital
                    queried_states.append(state['istate'] - 1)
                if (state['l']==3) and 'f' in orbital_query:         #f orbital
                    queried_states.append(state['istate'] - 1)
                
        return queried_states


    def plot_eigen(self, ax, size=20, cmap=None, cmap2=None,color='r', color_2='b',path_kpoints=[], label_1='', label_2='',
                   selected_orbitals=[], selected_orbitals_2=[],bandmin=0,bandmax=None,alpha=1,size_projection=False,y_offset=0.0,marker='o'):
        """ 
        Plot the band structure. The size of the points is the weigth of the selected orbitals.

        Options:

            (a) Relative weight between two compositions: selected_orbitals and selected_orbitas_2
                Format >>> selected_orbitals = [0,2,4]
            (b) Colormap enters as a string

        Arguments for plot layout:
        - size:     size of markers in scatterplot        [default 20]
        - alpha:    alpha value of markers in scatterplot [default 1]
        - label_1, label_2: plot labels [default empty string]

        Under development to include also colormap and a dictionary for the
        selection of the orbitals...
        example usage to get: 
             state #   2: atom   1 (Li ), wfc  2 (l=1 m= 1)
        
        plot_eigen(ax, path_kpoints=path_kpoints, selected_orbitals=[1], color=color, size=dotsize) 

            notice python counting; state# - 1 = selected_orbital index
        """
        #from numpy import arange # redundent
        import matplotlib.pyplot as plt
        import matplotlib as mpl
        if path_kpoints:
            if isinstance(path_kpoints,Path):
                path_kpoints = path_kpoints.get_indexes()
        if bandmax is None or bandmax > self.nbands:
            bandmax = self.nbands

        #Colormap
        if cmap:  color_map = plt.get_cmap(cmap)
        else:     color_map = plt.get_cmap('rainbow')
        if cmap2: color_map2 = plt.get_cmap(cmap2)
        else:     color_map2 = plt.get_cmap('rainbow')

        #get kpoint_dists
        kpoints_dists = calculate_distances(self.kpoints[:self.nkpoints])
 
        #make K-points labels
        ticks, labels = list(zip(*path_kpoints))
        ax.set_xticks([kpoints_dists[t] for t in ticks])
        ax.set_xticklabels(labels)
        ax.set_ylabel('E (eV)')

        #plot vertical lines
        for t in ticks: ax.axvline(kpoints_dists[t],c='gray',lw=0.8)
        ax.axhline(0,c='gray')
     
        # Plot bands for fixed size in a colormap
        if selected_orbitals_2:
           # No spin or full spinor
           if self.spin_components == 1 or self.spin_components == 4:
              w_rel = self.get_relative_weight(selected_orbitals=selected_orbitals, selected_orbitals_2=selected_orbitals_2)
              for ib in range(bandmin,bandmax):
                  eig = self.eigen[:,ib] + y_offset
                  eig_last = self.eigen[:,-1] + y_offset
                  if size_projection==True:
                     cax = ax.scatter(kpoints_dists,eig,s=size[:,ib],c=w_rel[:,ib],cmap=color_map,vmin=0,vmax=1,edgecolors='none',label=label_1,rasterized=True,zorder=2,marker=marker)
                  else:
                     cax = ax.scatter(kpoints_dists,eig,s=size,c=w_rel[:,ib],cmap=color_map,vmin=0,vmax=1,edgecolors='none',label=label_1,rasterized=True,zorder=2)

           # Spin polarized no SOC
           if self.spin_components == 2:
              w_rel1, w_rel2 = self.get_relative_weight(selected_orbitals=selected_orbitals, selected_orbitals_2=selected_orbitals_2)
              for ib in range(bandmin,bandmax):
                  eig1 = self.eigen1[:,ib] + y_offset
                  eig2 = self.eigen2[:,ib] + y_offset
                  if size_projection==True:
                     cax = ax.scatter(kpoints_dists,eig,s=size[:,ib],c=w_rel[:,ib],cmap=color_map,vmin=0,vmax=1,edgecolors='none',label=label_1,rasterized=True,zorder=2,marker=marker)
                  else:
                     cax = ax.scatter(kpoints_dists,eig1,s=size,c=w_rel1[:,ib],cmap=color_map,vmin=0,vmax=1,edgecolors='none',label=label_1,rasterized=True,zorder=2,marker=marker)
                     cax2= ax.scatter(kpoints_dists,eig2,s=size,c=w_rel2[:,ib],cmap=color_map2,vmin=0,vmax=1,edgecolors='none',label=label_1,rasterized=True,zorder=2,marker=marker)

        # Plot bands with changing size and a fixed color
        else:
            if self.spin_components == 1 or self.spin_components == 4:
               w_proj = self.get_weights(selected_orbitals=selected_orbitals)
               ib_max = np.where(w_proj==np.max(w_proj))[1][0] # needed for label
               for ib in range(bandmin,bandmax):
                   if ib==ib_max: lab = label_1
                   else:          lab = '_'+str(label_1)
                   eig = self.eigen[:,ib] + y_offset
                   cax = ax.scatter(kpoints_dists,eig,s=w_proj[:,ib]*size,c=color,edgecolors='none',alpha=alpha,label=lab,rasterized=True,zorder=2,marker=marker)

            elif self.spin_components == 2:
                 w_proj1, w_proj2 = self.get_weights(selected_orbitals=selected_orbitals)
                 ib_max1, ib_max2 = np.where(w_proj1==np.max(w_proj1))[1][0], np.where(w_proj2==np.max(w_proj2))[1][0]
                 for ib in range(bandmin,bandmax):
                     lab1, lab2 = ['_'+label_1,'_'+label_2]
                     if ib==ib_max1: lab1 = label_1
                     if ib==ib_max2: lab2 = label_2
                     eig1, eig2 = self.eigen1[:,ib], self.eigen2[:,ib]
                     cax = ax.scatter(kpoints_dists,eig1,s=w_proj1[:,ib]*size,c=color  ,edgecolors='none',alpha=alpha,label=lab1,marker=marker)
                     cax2= ax.scatter(kpoints_dists,eig2,s=w_proj2[:,ib]*size,c=color_2,edgecolors='none',alpha=alpha,label=lab2,marker=marker)

        ax.set_xlim(0, max(kpoints_dists))
        return cax
    
    def get_weights(self,selected_orbitals=[],bandmin=0,bandmax=None):
        if bandmax is None:
           bandmax = self.nbands

        if self.spin_components == 1:

           # Selection of the bands
           w_proj = zeros([self.nkpoints,self.nbands])
           for ik in range(self.nkpoints):
               for ib in range(bandmin,bandmax):
                   w_proj[ik,ib] = sum(abs(self.proj[ik,selected_orbitals,ib])**2)
           return w_proj

        if self.spin_components == 2:

           # Selection of the bands
           w_proj1 = zeros([self.nkpoints,self.nbands])
           w_proj2 = zeros([self.nkpoints,self.nbands])

           for ik in range(self.nkpoints):
               for ib in range(bandmin,bandmax):
                   w_proj1[ik,ib] = sum(abs(self.proj1[ik,selected_orbitals,ib])**2)
                   w_proj2[ik,ib] = sum(abs(self.proj2[ik,selected_orbitals,ib])**2)
           return w_proj1, w_proj2

    def get_dorbitals_projection(self,selected_orbitals=[],bandmin=0,bandmax=None):
        """
        This function return the weights for d-orbitals in the basis of a1g, e+
        and e- 
        selected_orbitals must the indices of the d-orbital in the array in the order of QE
        """
        if bandmax is None:
           bandmax = self.nbands

        #if self.spin_components == 1:

           # Selection of the bands
           #w_proj = zeros([self.nkpoints,self.nbands])
           #for ik in range(self.nkpoints):
           #    for ib in range(bandmin,bandmax):
           #        w_proj[ik,ib] = sum(abs(self.proj[ik,selected_orbitals,ib])**2)
           #return w_proj

        if self.spin_components == 2:

           # Selection of the bands
           w_proj1 = zeros([self.nkpoints,self.nbands])
           w_proj2 = zeros([self.nkpoints,self.nbands])

           for ik in range(self.nkpoints):
               for ib in range(bandmin,bandmax):

                   w_proj1[ik,ib] = sum(abs(self.proj1[ik,selected_orbitals,ib])**2)
                   w_proj2[ik,ib] = sum(abs(self.proj2[ik,selected_orbitals,ib])**2)
           return w_proj1, w_proj2

    def get_pdos(self,selected_orbitals=None,bandmin=0,bandmax=None,energy_steps=100,e_min=-10.0,e_max=5.0,Gamma=0.1):

        print(selected_orbitals)

        energy_grid = arange(e_min,e_max,(e_max-e_min)/energy_steps)
        if bandmax is None:
           bandmax = self.nbands
        if selected_orbitals is None: selected_orbitals = range(self.nproj)

        if self.spin_components == 2:
           pdos_up, pdos_dw = zeros([energy_steps,self.nproj]), zeros([energy_steps,self.nproj])
           for ik in range(self.nkpoints):
               for ib in range(bandmin,bandmax):
                   for io in selected_orbitals:
                       for ie,e in enumerate(energy_grid):
                           pdos_up[ie,io] = pdos_up[ie,io] + abs(conjugate(self.proj1[ik,io,ib])*self.proj1[ik,io,ib])*self._lorentz(self.eigen1[ik,ib],e,Gamma)

        return energy_grid, pdos_up #, dos_up       

    def _lorentz(self,x,x_0,Gamma):
        
        return (1.0/pi)*(0.5*Gamma)/((x-x_0)**2 + (0.5*Gamma)**2)
    
    def get_relative_weight(self,selected_orbitals=[],selected_orbitals_2=[],bandmin=0,bandmax=None):
        if bandmax is None:
            bandmax = self.nbands

        # No spin polarized
        if self.spin_components == 1 or self.spin_components == 4:
           # Selection of the bands
           w_rel = zeros([self.nkpoints,self.nbands])
           for ik in range(self.nkpoints):
               for ib in range(bandmin,bandmax):
                   a = sum(abs(self.proj[ik,selected_orbitals  ,ib])**2)
                   b = sum(abs(self.proj[ik,selected_orbitals_2,ib])**2)
                   w_rel[ik,ib] = a/(a+b)
           return w_rel

        # Spin polarized collinear
        if self.spin_components == 2:
           # Selection of the bands
           w_rel1 = zeros([self.nkpoints,self.nbands])
           w_rel2 = zeros([self.nkpoints,self.nbands])
           for ik in range(self.nkpoints):
               for ib in range(bandmin,bandmax):
                   a1 = sum(abs(self.proj1[ik,selected_orbitals  ,ib])**2)
                   b1 = sum(abs(self.proj1[ik,selected_orbitals_2,ib])**2)
                   w_rel1[ik,ib] = a1/(a1+b1)
                   a2 = sum(abs(self.proj2[ik,selected_orbitals  ,ib])**2)
                   b2 = sum(abs(self.proj2[ik,selected_orbitals_2,ib])**2)
                   w_rel2[ik,ib] = a2/(a2+b2)
           return w_rel1, w_rel2

    def get_kpoints(self):
        """ Return kpoints
        """
        kpoints = []
        datafile_xml = self.datafile_xml
        for word in self.datafile_xml.findall("EIGENSTATES/K-POINT"):
            kpoints.append( list( map(float, word.text.split()) )  )
        
        return kpoints

    def get_eigen(self):
        """ Return eigenvalues
        """
        datafile_xml = self.datafile_xml
        eigen = []
        eigen1 = []
        eigen2 = []

        # No spin polarized
        if self.spin_components == 1 or self.spin_components == 4:

            eigen =  [ list( map(float, word.text.split())) for word in self.datafile_xml.findall("EIGENSTATES/E") ] 

            self.eigen = np.array(eigen)*RytoeV - self.fermi
            return self.eigen

        # Spin polarized
        if self.spin_components == 2:
           
            eigen_prov =  [ list( map(float, word.text.split())) for word in self.datafile_xml.findall("EIGENSTATES/E") ] 
            eigen_aux = np.array(eigen_prov)*RytoeV 
            self.eigen1 = eigen_aux[            0:  self.nkpoints,:] - self.fermi
            self.eigen2 = eigen_aux[self.nkpoints:2*self.nkpoints,:] - self.fermi

            return self.eigen1, self.eigen2

    def write_proj(self,filename='proj'):
        """
        Write the projection array in a numpy file
        """
        np.savez(filename,proj=self.proj,weights=self.weights)
            
    def get_proj(self):
        """ Return projections
        """
        datafile_xml = self.datafile_xml

        if self.spin_components == 1 or self.spin_components == 4:
            proj  = zeros([self.nkpoints,self.nproj,self.nbands],dtype=complex)

            data_atomic_wfc = self.datafile_xml.findall("EIGENSTATES/PROJS/ATOMIC_WFC")
            for ik in range(self.nkpoints):
                for ip in range(self.nproj):
                    i_data = ik*self.nproj + ip
                    projlist = data_atomic_wfc[i_data].text.splitlines()[1:-1]
                    atom_aux = []
                    for c in projlist:
                        z = float(c.split()[0]) + 1.0j*float(c.split()[1])
                        atom_aux.append(z)
                    proj[ik,ip] = atom_aux
            
            self.proj = np.array(proj)

            return proj

        # Spin polarized
        elif self.spin_components == 2:
            print('return spin polarized projections')
           
            # Two independent spinors
            data_atomic_wfc = self.datafile_xml.findall("EIGENSTATES/PROJS/ATOMIC_WFC")
            proj1 = zeros([self.nkpoints,self.nproj,self.nbands],dtype=complex)
            proj2 = zeros([self.nkpoints,self.nproj,self.nbands],dtype=complex)
            
            for ik in range(self.nkpoints):
                for ip in range(self.nproj):
                    i_data1 = ik*self.nproj + ip
                    i_data2 = (ik+self.nkpoints)*self.nproj + ip
                    projlist1 = data_atomic_wfc[i_data1].text.splitlines()[1:-1]
                    projlist2 = data_atomic_wfc[i_data2].text.splitlines()[1:-1]
                    atom_aux1, atom_aux2 = [], []
                    for c in projlist1:
                        z = float(c.split()[0]) + 1.0j*float(c.split()[1])
                        atom_aux1.append(z)
                    for c in projlist2:
                        z = float(c.split()[0]) + 1.0j*float(c.split()[1])
                        atom_aux2.append(z)
                    proj1[ik,ip] = atom_aux1
                    proj2[ik,ip] = atom_aux2

            self.proj1 = np.array(proj1)
            self.proj2 = np.array(proj2)

            return proj1, proj2
       
    def shift_bands(self,qpcorrection,vb,cb):
        """
        Shift band structure, e.g. to account for a G0W0 run.

        The idea is to shift the bands by the qp corrections in order to use 
        the weights from a projwfc calculation.

        Note that the path used in QE must be the same as the one used in Yambo

        - vb and cb allow for tuning the bands that are to be corrected
        """

        for ib in  range(self.nbands-vb,self.bands,self.nbands+cb): 
            self.eigen[:,ib] = self.eigen[:,ib]+qpcorrection[:,ib]
 
    def add_scissor(self,n_val,scissor):
        """
        Apply a scissor operator to the bands, e.g. to account for a G0W0 run.
            - n_val: number of valence bands
            - scissor: list of three numbers [shift, cond_stretch, val_stretch]
            - bands: array of eigenvalues
        """
        def band_formats(bands):
            """ Yambo format has leading spin axis
            """
            if len(bands.shape)>2: return bands[0] #Yambo spin-polarized case not supported
            else:                  return bands

        scissored_bands = np.zeros(self.eigen.shape)

        tmp_bands    = band_formats(self.eigen)
        tmp_sc_bands = band_formats(scissored_bands)

        top_v, bottom_c = tmp_bands[:,n_val-1], tmp_bands[:,n_val]
        ind_k_dir_gap = np.argmin(bottom_c-top_v)
        ev_max, ec_min = top_v[ind_k_dir_gap], bottom_c[ind_k_dir_gap]

        for ik in range( self.nkpoints ):
            for ib in range( self.nbands ):
                if ib<n_val: tmp_sc_bands[ik][ib] = ev_max-(ev_max-tmp_bands[ik][ib])*scissor[2]
                else: tmp_sc_bands[ik][ib] = ec_min+scissor[0]+(tmp_bands[ik][ib]-ec_min)*scissor[1]
        self.eigen0 = self.eigen # save original bands
        self.eigen  = scissored_bands

    def __str__(self):
        version_number = float(re.findall(r"([0-9.]+)",self.qe_version)[0])
        lines = []; app = lines.append
        app(marquee(self.__class__.__name__))

        app(f"Running projwfcxml for QE version {self.qe_version}")
        app(f"nkpoints: {self.nkpoints}")
        app(f"nbands:   {self.nbands}")
        if version_number>=7: #replace with order_l-j-m-mj and order_j-l-m-mj 
            for n,state in enumerate(self.states):
                app(f"n: {n} -> iatom:{state['iatom']} atype:{state['atype']} wfc:{state['wfc']} l:{state['l']} j:{state['j']} m_j:{state['m_j']}")
            return "\n".join(lines)
        else:
            for n,state in enumerate(self.states):
                app(f"n: {n} -> iatom:{state['iatom']} atype:{state['atype']} wfc:{state['wfc']} j:{state['j']} l:{state['l']} m:{state['m']} m_j:{state['m_j']}")
            return "\n".join(lines)



# Author: Davide Romanin (revised: FP, RR)
#
# This file is part of the yambopy project
#
import os
from netCDF4 import Dataset
import numpy as np
from itertools import product
from yambopy import YamboLatticeDB
from yambopy.tools.string import marquee
from yambopy.units import I

class YamboBSEKernelDB(object):
    """ Read the BSE Kernel database from yambo.
        It reads <t1| K |t2> where K is the kernel and t_i transition indices.
        
        Can only be used if yambo is run with parallel IO.
        
        Only supports "RESONANT" case for BSE calculation.
        TODO: support more cases
    """
    def __init__(self,lattice,kernel):
        if not isinstance(lattice,YamboLatticeDB):
            raise ValueError('Invalid type for lattice argument. It must be YamboLatticeDB')
        
        self.lattice  = lattice
        self.kernel   = kernel

    @classmethod
    def from_db_file(cls, lattice, Qpt=1, folder='.'):
        """Initialize this class from a ndb.BS_PAR_Q# file."""
        filename = 'ndb.BS_PAR_Q%d' % Qpt
        path_filename = os.path.join(folder, filename)
        if not os.path.isfile(path_filename):
            raise FileNotFoundError(f"File {path_filename} not found in YamboExcitonDB")

        with Dataset(path_filename) as database:
            if 'BSE_RESONANT' in database.variables:
                # Read as transposed since dimensions in netCDF are inverted
                reker, imker = database.variables['BSE_RESONANT'][:].T
                ker = reker + imker*I
                
                # Transform the triangular matrix to a square Hermitian matrix
                kernel = np.conjugate(np.transpose(np.triu(ker))) + np.triu(ker)
                kernel[np.diag_indices(len(kernel))] *= 0.5
                
                # Check if the kernel is Hermitian
                if not np.allclose(kernel, np.conjugate(kernel.T)):
                    raise ValueError("The constructed kernel matrix is not Hermitian")
            else:
                raise ValueError('Only BSE_RESONANT case supported so far')

        return cls(lattice, kernel)

    @property
    def ntransitions(self): return len(self.kernel)

    def consistency_BSE_BSK(self,excitons):
        """ Check that exciton and kernel dbs are consistent
        """
        if excitons.nexcitons != self.ntransitions:
            print('[WARNING] Exciton and transition spaces have different dimensions!')
        if excitons.ntransitions != self.ntransitions:
            print('[WARNING] Mismatch in ntransitions between ExcitonDB and BSEkernelDB!')        

    def get_kernel_exciton_basis(self,excitons):
        """ Switch from transition |tq>=|kc,k-qv> to excitonic |lq> basis. 
            In this basis the kernel is diagonal.
            
            <l|K|l> = sum_{t,t'}( <l|t><t|K|t'><t'|l> )
                    = sum_{t,t'}( (A^l_t)^* K_tt' A^l_t' ) 
       
            exciton: YamboExcitonDB object 
            Here t->kcv according to table from YamboExcitonDB database
        """
        kernel   = self.kernel
        Nstates  = self.ntransitions
        eivs     = excitons.eigenvectors
        self.consistency_BSE_BSK(excitons)

        # Basis transformation
        kernel_exc_basis  = np.einsum('ij,kj,ki->k', kernel, eivs, np.conj(eivs), optimize=True)
        #kernel_exc_basis = np.zeros(Nstates,dtype=complex)
        #for il in range(Nstates):
        #    kernel_exc_basis[il] = np.dot( np.conj(eivs[il]), np.dot(kernel,eivs[il]) )

        return kernel_exc_basis

    def get_kernel_value_bands(self,excitons,bands):
        """ Get value of kernel matrix elements 
            as a function of k in BZ for fixed c,v bands:
            
                K_cv(k,p) = <ck,vk-q|K|cp,vp-q>
                
            exciton: YamboExcitonDB object
            bands = [iv,ic] (NB: enumerated starting from one instead of zero) 
        """
        table  = excitons.table
        nk     = self.lattice.nkpoints
        kernel = self.kernel
        self.consistency_BSE_BSK(excitons)
        
        if bands[0] not in table[:,1] or bands[1] not in table[:,2]:
            raise ValueError('Band indices not matching available transitions')
             
        # Wcv defined on the full BZ (only a subset will be filled)
        Wcv = np.zeros((nk,nk),dtype=complex)
        # Find indices where selected valence band appears
        t_v = np.where(table[:,1]==bands[0])[0]
        # Find indices where selected conduction band appears
        t_c = np.where(table[:,2]==bands[1])[0]
        # Among those, find subset of indices where both appear together
        t_vc = [t for t in t_v if t in t_c ]

        # Iterate only on the subset
        for it1_subset, it2_subset in product(t_vc,repeat=2):
            ik = table[it1_subset][0]
            ip = table[it2_subset][0]
            Wcv[ik-1,ip-1] = kernel[it1_subset,it2_subset]
        return Wcv

    def get_string(self,mark="="):
        lines = []; app = lines.append
        app( marquee(self.__class__.__name__,mark=mark) )
        app( "kernel mode: RESONANT" )
        app( "number of transitions: %d"%self.ntransitions )
        return '\n'.join(lines)
    
    def __str__(self):
        return self.get_string()



# Copyright (C) 2018 Henrique Pereira Coutada Miranda
# All rights reserved.
#
# This file is part of yambopy
#
import numpy as np
import unittest
import os
from yambopy.dbs.excitondb import YamboExcitonDB
from yambopy.dbs.latticedb import YamboLatticeDB
from yambopy.dbs.electronsdb import YamboElectronsDB
from qepy.lattice import Path

test_path = os.path.join(os.path.dirname(__file__),'..','..','data','refs','bse')

class TestYamboExcitonDB(unittest.TestCase):

    def test_yamboexcitondb(self):

        #define path in reduced coordinates
        npoints = 20
        path = Path([ [[0.0, 0.0, 0.0],'$\\Gamma$'],
                      [[0.5, 0.0, 0.0],'M'],
                      [[1./3,1./3,0.0],'K'],
                      [[0.0, 0.0, 0.0],'$\\Gamma$']],
                      [int(npoints*2),int(npoints),int(np.sqrt(5)*npoints)])

        #load databases
        lat  = YamboLatticeDB.from_db_file(os.path.join(test_path,'SAVE','ns.db1'))
        exc  = YamboExcitonDB.from_db_file(lat,folder=os.path.join(test_path,'yambo'))
        electrons = YamboElectronsDB(lat,save=os.path.join(test_path,'SAVE')) 

        #show output
        sort_e, sort_i = exc.get_sorted()
        exc.get_degenerate(1)

        #write exc_I.dat and exc_E.dat
        exc.write_sorted('exc')

        #calculate chi
        w,chi = exc.get_chi()

        #load reference
        eps_filename = os.path.join(test_path,'o-yambo.eps_q1_diago_bse')
        w_ref,chi_imag_ref,chi_real_ref = np.loadtxt(eps_filename,unpack=True)[:3]

        #get amplitude and phases
        kpoints, amplitude, phase = exc.get_amplitudes_phases((1,0,))

        #exciton_bandstructure
        exc.plot_exciton_bs(electrons, path, (1,2,), args_plot={'c':'g'},space='bands',show=False)

    def tearDown(self):
        if os.path.isfile('exc_I.dat'): os.remove('exc_I.dat')
        if os.path.isfile('exc_E.dat'): os.remove('exc_E.dat')


if __name__ == '__main__':
    unittest.main()



from yambopy import *
from schedulerpy import *
import time
import os
from copy import deepcopy
overflow = 1e8

def integerize(number):
    """
    Check if number is integer, if False make it integer
    """
    if isinstance(number,int): return number
    if isinstance(number,float):
        if number.is_integer(): return int(number)
        else: return integerize(number*10)

class YamboRTStep_Optimize():
    """ 
    Class to run convergence tests for the RT time step.

    Note: time steps must be given in as units.    

    - Needs an initialised RT SAVE
    - Needs an RT input 
    - Optional arguments: directory paths, max time step, time step increase, max number of runs
    - Optional: specify "yscheduler" as an instance of schedulerpy to run on clusters. This needs dynamical managing of submitted jobs.
    - Optional: add convergence loop with respect to field intensity

    Example of use:
        .. code-block:: python
    
            YamboRTStep_Optimize(input_path,SAVE_path,RUN_path,ref_time,TStep_MAX,TStep_increase,NSimulations)
    
        .. code-block:: python
    
            YamboRTStep_Optimize(input_path,SAVE_path,RUN_path,ref_time,TStep_MAX,TStep_increase,NSimulations,FieldInt)

    """

    def __init__(self,input_path='./yambo.in',SAVE_path='./SAVE',RUN_path='./RT_time-step_optimize',yambo_rt='yambo_rt',\
                 ref_time=60,TStep_MAX=30,TStep_increase=5,NSimulations=6,FieldInt=None,yscheduler=None,\
                 tol_eh=1e-4,tol_pol=5e-3,Tpoints_min=30,plot_results=True):
        #Configuring schedulers
        self.frontend = Scheduler.factory(scheduler="bash")
        if yscheduler is not None: #Here we use, e.g., slurm 
            self.jobrun = yscheduler
            self.wait_up = True #This will enable calls to a function that will make the code wait upon completion of previous submitted job
        else: 
            self.jobrun = Scheduler.factory(scheduler="bash")
            self.wait_up = False
        #Setting global variables
        input_path, input_name = input_path.rsplit('/',1)
        self.yin = YamboIn.from_file(filename=input_name,folder=input_path)
        self.RUN_path = RUN_path
        self.yambo_rt = yambo_rt

        self.ref_time = ref_time #Simulation duration (fs) after field ends.
        self.TStep_MAX = TStep_MAX
        self.TStep_increase = TStep_increase
        self.NSimulations = NSimulations
        self.Tpoints_min = Tpoints_min
        self.tol_eh = tol_eh
        self.tol_pol= tol_pol
        self.FieldInt = FieldInt
        self.time_odm = 1 # Important: this needs to be an integer
        #Generate directories
        self.create_folder_structure(SAVE_path)
        #Start IO
        if self.wait_up: prnt_io = False
        else: prnt_io = True
        self.yf = YamboIO(out_name='YAMBOPY_RTStepConvergence.log',out_path=self.RUN_path,print_to_shell=prnt_io)
        self.yf.IO_start()
        #
        if self.wait_up: self.yf.msg("The workflow is run using job submission through a scheduler.")
        else: self.yf.msg("The workflow is run locally.")
        #Check for consistent input parameters
        if integerize(self.TStep_MAX) % integerize(self.TStep_increase) !=0: #Here RaiseError may be used
            self.yf.msg("The polarization is computed at discrete times.")
            self.yf.msg("In order to compare efficiently results with different time steps,")
            self.yf.msg("please select a time increment that divides exactly the max time step.")
            self.yf.msg("Exiting...")
            exit() 
        if self.NSimulations < 2:
            self.yf.msg("NSimulations is too small to perform convergence tests.")
            self.yf.msg("Exiting...")
            exit()
        #Compute the dipoles, then prepare RT input
        self.FIND_values()
        self.COMPUTE_dipoles()
        #Run RT simulations and analyse data
        self.RUN_convergence()
        #[OPTIONAL] plot results
        if plot_results: self.PLOT_output()

        self.yf.IO_close()

    def create_folder_structure(self,SAVE_path):
        
        if not os.path.isdir(self.RUN_path):
            shell = self.frontend
            shell.add_command('mkdir -p %s'%self.RUN_path)
            shell.add_command('cd %s ; ln -s ../%s . ; cd ..'%(self.RUN_path,SAVE_path))
            shell.run()
            shell.clean()

        if not os.path.islink('%s/SAVE'%self.RUN_path):
            shell = self.frontend
            shell.add_command('cd %s ; ln -s ../%s . ; cd ..'%(self.RUN_path,SAVE_path))
            shell.run()
            shell.clean()

    def FIND_values(self):
        """ 
        Determine time step values to be run and simulation lengths.
        """

        #Check which laser is used
        if self.yin['Field1_kind']=="DELTA":
            self.yf.msg("Field kind: DELTA")
            FieldTime = 0.

        if self.yin['Field1_kind']=="QSSIN":
            self.yf.msg("Field kind: QSSIN")
            if 'Field1_FWHM' in self.yin.variables.keys():
                if self.yin['Field1_FWHM']==0.: # Here RaiseError may be used
                    self.yf.msg("Please use the variable Field1_FWHM to set field width (not Field1_kind)")
                    self.yf.msg("Exiting...")
                    exit()
            else:
                self.yf.msg("Please use the variable Field1_FWHM to set field width (not Field1_Width)")
                self.yf.msg("Exiting...")
                exit()
            fwhm = self.yin['Field1_FWHM'][0]
            sigma = fwhm/(2*np.sqrt(2*np.log(2))) #Standard deviation from FWHM from normal distr.
            sigma = np.around(sigma,2)
            self.yf.msg("with FWHM:    %f %s"%(fwhm,self.yin['Field1_FWHM'][1]))
            self.yf.msg("with dev.st.: %f %s"%(sigma,self.yin['Field1_FWHM'][1]))
            FieldTime = 6.*sigma      

        self.yf.msg("Field direction: %s"%(str(self.yin['Field1_Dir'][0])))

        #Set field intensity if given in input
        if self.FieldInt is not None: self.yin['Field1_Int']=[ self.FieldInt, 'kWLm2' ]
        else: self.FieldInt=self.yin['Field1_Int'][0] 

        #Set time steps
        time_steps = [ self.TStep_MAX - i*self.TStep_increase for i in range(self.NSimulations)]
        self.time_steps = np.array( [ ts for ts in time_steps if ts>0 ] )
        self.NSimulations = len(self.time_steps)
        self.TSteps_min_max=[self.TStep_MAX,self.TStep_MAX-(self.NSimulations-1)*self.TStep_increase]
        # We have to consider the case of non-integer below-as time steps
        if self.TStep_increase % 1 > 0: self.time_odm = 1000

        #Set simulations time settings (field time + lcm(time_steps) + hardcoded duration to analyse)
        ts_integerized = np.array( [ int('{0:g}'.format(ts*self.time_odm)) for ts in self.time_steps ] ) 
        # Grievous floating-point issues here for non-integer time steps... solved with above line
        ts_lcm = float(np.lcm.reduce(ts_integerized))/(1000.*self.time_odm) # in fs
        if self.ref_time/ts_lcm<self.Tpoints_min:
            self.yf.msg("[ERR] less than %d time points for polarization."%self.Tpoints_min)
            self.yf.msg("Exiting...")
            exit()
        self.yin['Field1_Tstart'] = [ts_lcm, 'fs']
        NETime = ts_lcm + FieldTime + self.ref_time
        self.yin['NETime'] = [ NETime, 'fs' ]
        self.NETime = NETime
        self.yf.msg("Total duration of simulations set to: %f fs"%NETime)
        self.yin['IOCachetime'] = [[ts_lcm,5*ts_lcm],'fs']

    def COMPUTE_dipoles(self,DIP_folder='dipoles'):
        """
        Compute the dipoles once and for all:
        In order for the dipoles to be compatible with a negf run 
        [a default optics run does not produce compatible dipoles], 
        the 'negf' argument is appended which causes the calculation to crash
        *after* the dipoles are computed.
        """
        if not os.path.isfile('%s/%s/ndb.dipoles'%(self.RUN_path,DIP_folder)):
            # Input...
            ydipoles = YamboIn()
            ydipoles.arguments.append('dipoles')
            ydipoles.arguments.append('negf')
            ydipoles['DIP_ROLEs'] = self.yin['DIP_ROLEs']
            ydipoles['DIP_CPU'] = self.yin['DIP_CPU']
            ydipoles['DipBands'] = self.yin['DipBands']
            ydipoles.write('%s/dipoles.in'%self.RUN_path)
            # Running...
            self.yf.msg("Running dipoles...")
            shell = deepcopy(self.jobrun)
            shell.name = 'dipoles'
            shell.add_mpirun_command('%s -F dipoles.in -J %s -C %s 2> %s.log'%(self.yambo_rt,DIP_folder,DIP_folder,DIP_folder))
            shell.run(filename='%s/rt.sh'%self.RUN_path)
            if self.wait_up: wait_for_job(shell,self.RUN_path) #OLD VERSION: self.wait_for_job(shell)
            shell.clean() 
        else:
            self.yf.msg("Dipoles found.")

        self.DIP_folder = DIP_folder

    def input_to_run(self,param,value,units):
        """
        Generate input for a specific run
        """
        yrun = deepcopy(self.yin)
        yrun[param] = [ value, units]
        return yrun

    def check_if_run_has_been_already_done(self,folder):
        """
        Skips running ts if:
        - ndb.output_polarization is found (sloppy check), AND
        - The string "Game over" appears in the report file (strict check)
        """
        skip1=False
        skip2=False
        ndb_pol_file='%s/%s/ndb.output_polarization'%(self.RUN_path,folder)
        report_file='%s/%s/r-%s_dipoles_negf'%(self.RUN_path,folder,folder)

        if os.path.isfile(ndb_pol_file): skip1=True
        if os.path.isfile(report_file):
            with open(report_file) as report:
                if 'Game Over' in report.read(): skip2=True

        return skip1*skip2

    def RUN_convergence(self,param='RTstep',units='as'):
        """
        Run the yambo_rt calculations flow.
        """        
        self.yf.msg("Running RT time step convergence...")
        RToutput  =    []
        NaN_check =    []
        eh_check  =    []
        pol_sq_check = []
        pol_x_check  = []
        time_steps = self.time_steps
        for i,ts in enumerate(time_steps):
            self.yf.msg("Running simulation for time step: %s %s"%('{0:g}'.format(ts),units))

            # Part 1: file preparation and run
            if self.time_odm==1: filename = '%s_%05d%s.in'%(param,ts,units)
            if self.time_odm==1000: filename = '%s_%05d%s.in'%(param,ts*self.time_odm,'zs')
            folder   = filename.split('.')[0]
            #Skip execution if output found:
            if self.check_if_run_has_been_already_done(folder):
                self.yf.msg("Found output for time step: %s %s"%('{0:g}'.format(ts),units))
            else:
                yrun = self.input_to_run(param,float(ts),units)
                yrun.write('%s/%s'%(self.RUN_path,filename))
                shell = deepcopy(self.jobrun)
                shell.name = '%s_%s_%s'%('{:.0E}'.format(self.FieldInt).replace("E+0", "E"),'{0:g}'.format(ts),shell.name)
                shell.add_mpirun_command('%s -F %s -J %s,%s -C %s 2> %s.log'%(self.yambo_rt,filename,folder,self.DIP_folder,folder,folder))
                shell.run(filename='%s/rt.sh'%self.RUN_path)
                if self.wait_up: wait_for_job(shell,self.RUN_path) #OLD VERSION: self.wait_for_job(shell)
                shell.clean()

            # Part 2: perform single-run analysis and store output
            out_dir = '%s/%s'%(self.RUN_path,folder)
            #Read output
            RTDB = YamboRTDB(calc=out_dir) #Read output
            RToutput_no_nan, NaN_test = self.nan_test(RTDB)              #[TEST1] NaN and overflow
            RToutput.append(RToutput_no_nan)
            if NaN_test: eh_test = self.electron_conservation_test(RTDB) #[TEST2] Electron number
            else:        eh_test = False
            NaN_check.append(NaN_test) 
            eh_check.append(eh_test) 

            # Part 3: perform polarization tests between subsequent runs
            if i==0: passed_counter = 0
            if i>0: 
                pol_sq_test, pol_x_test, passed_counter = self.ANALYSE_pol(RToutput,eh_check,passed_counter) #[TEST3],[TEST4] Polarization squared and along field direction
                pol_sq_check.append(pol_sq_test)
                pol_x_check.append(pol_x_test)

            # Part 4: decide if convergence was reached or we have to keep going
            if passed_counter==2:
                TStep_passed = self.time_steps[i-1]
                break
            
        if passed_counter==2: self.TStep_passed = TStep_passed
        if passed_counter==1: self.TStep_passed = self.time_steps[-1]
        if passed_counter==0: self.TStep_passed = None
            
        self.NSimulations = len(RToutput)
        self.RToutput = RToutput
        self.ANALYSE_output(NaN_check,eh_check,pol_sq_check,pol_x_check,passed_counter)

    def ANALYSE_output(self,NaN,eh,pol2,polx,passed,units='as'):
        """
        Output information and suggestion for an optimal time step.
        - There are two values of tolerance: one for carriers, one for polarization
        - Four increasingly stringent checks are performed:
            [1] NaN and overflow check to exclude botched runs
            [2] Conservation of electron number check
            [3] Error check of |pol|^2 (assuming lowest time step as reference)
            [4] Error check of pol along the field direction
        """
        self.yf.msg("---------- ANALYSIS ----------")

        NaN_passed = sum(NaN)
        self.yf.msg("[1] NaN and overflow test:")
        self.yf.msg("    Passed by %d out of %d."%(NaN_passed,self.NSimulations))

        eh_passed = sum(eh)
        self.yf.msg("[2] Conservation of electron number test (tol=%.0e):"%self.tol_eh)
        self.yf.msg("    Passed by %d out of %d."%(eh_passed,self.NSimulations))

        pol2_passed = sum(pol2)
        self.yf.msg("[3] Error in |pol|^2 test (tol=%.0e):"%self.tol_pol)
        self.yf.msg("    Passed by %d out of %d."%(pol2_passed,self.NSimulations-1))

        polx_passed = sum(polx)
        self.yf.msg("[4] Error in pol along field test (tol=%.0e):"%self.tol_pol)
        self.yf.msg("    Passed by %d out of %d."%(polx_passed,self.NSimulations-1))

        if passed == 1:
            self.yf.msg(" ")
            self.yf.msg("[WARNING] The lowest time step passed all the tests, but the")
            self.yf.msg("          additional safety run with a reduced step was")
            self.yf.msg("          not done due to NSimulations limit being reached.")

        if self.NSimulations == 2 or self.NSimulations == 3:
            self.yf.msg(" ")
            self.yf.msg("[WARNING] The largest time step already looks converged.")

        tp = self.TStep_passed
        self.yf.msg(" ")
        self.yf.msg("Based on the analysis, the suggested time step is: ")
        if tp is not None: self.yf.msg("### %s %s ###"%('{0:g}'.format(tp),units))
        else: self.yf.msg("[ERR] NSimulations limit reached before converged value was found.")
        self.yf.msg("------------------------------")        
 
    def ANALYSE_pol(self,RToutput,eh_check,passed):
        """
        Driver with the logical structure to manage polarization tests
        """
        if eh_check[-1]==True and eh_check[-2]==True:
            pol_sq_test = self.pol_error_test(RToutput,which_pol='pol_sq')
            pol_x_test  = self.pol_error_test(RToutput,which_pol='pol_along_field')
            
            if pol_sq_test and pol_x_test: passed = passed + 1
            else: passed = 0

        else: 
            pol_sq_test = False
            pol_x_test  = False

        return pol_sq_test, pol_x_test, passed

    def nan_test(self,RTDB):
        """ 
        Check computed polarizations for NaN values.
        """
        NaN_test = True
        # Check for NaN
        if np.isnan(RTDB.polarization).any() or np.isnan(RTDB.diff_carriers).any(): 
            RTDB.polarization = np.nan_to_num(RTDB.polarization) #Set to zero for plots
            NaN_test = False 
            #self.yf.msg("[WARNING] Yambo produced NaN values during this run")
        # Check for +/-Infinity
        if np.greater(np.abs(RTDB.polarization),overflow).any():
            RTDB.polarization[np.abs(RTDB.polarization)>overflow] = 0. #Set to zero for plots
            NaN_test = False
            #self.yf.msg("[WARNING] Yambo produced Infinity values during this run")           
 
        return RTDB, NaN_test

    def electron_conservation_test(self,RTDB):
        """
        Tests if elements of ratio_carriers are greater than tolerance.
        If any of them is, then the simulation in question has not passed the eh_test.
        """
        eh_carriers = np.greater(np.abs(RTDB.ratio_carriers),self.tol_eh)
        if any(eh_carriers): eh_test = False
        else:                eh_test = True
        return eh_test

    def pol_error_test(self,RTout,which_pol):
        """
        Computes the relative errors of the polarizations for each cached time.
        The cached times coincide for different runs.
        """
        pol_analyse= []
        pol_n1 = RTout[-1].polarization   
        pol_n0 = RTout[-2].polarization 
        if which_pol == 'pol_sq':  # Test for |pol|^2
            pol_analyse_n1 = pol_n1[0]*pol_n1[0] + pol_n1[1]*pol_n1[1] + pol_n1[2]*pol_n1[2] 
            pol_analyse_n0 = pol_n0[0]*pol_n0[0] + pol_n0[1]*pol_n0[1] + pol_n0[2]*pol_n0[2] 
        if which_pol == 'pol_along_field': # Test for pol along field
            pol_analyse_n1, _ = self.pol_along_field(pol_n1)
            pol_analyse_n0, _ = self.pol_along_field(pol_n0)            
 
        #Perform the test
        rel_err_pol = (pol_analyse_n1-pol_analyse_n0)/self.tol_pol
        error = np.greater(rel_err_pol,1.).any()
        if error: pol_test = False
        else:     pol_test = True      

        return pol_test
 
    def pol_along_field(self,pol):
        field = self.yin['Field1_Dir'][0]
        n_dirs = np.count_nonzero(field)
        pol_x = field[0]*pol[0]+field[1]*pol[1]+field[2]*pol[2]
        pol_x = pol_x/float(n_dirs)
        if field[0]!=0.:   pol_label='pol-x'
        elif field[1]!=0.: pol_label='pol-y'
        elif field[2]!=0.: pol_label='pol-z'
        else:              pol_label='pol-xyz'
        return pol_x,pol_label

    def PLOT_output(self,save_dir='plots'):
        """
        Generic plots generated by default, to be accessed by the user
        """
        import matplotlib.pyplot as plt

        plots_dir = 'runs'
        odm = self.time_odm
        
        if odm==1: units='as'
        if odm==1000: units='zs'
        for ts in self.time_steps: plots_dir += '_%s'%('{0:g}'.format(ts*odm))

        self.yf.msg("Plotting results.")
        out_dir = '%s/%s'%(self.RUN_path,save_dir)
        shell = self.frontend
        if not os.path.isdir(out_dir): shell.add_command('mkdir -p %s'%out_dir)
        shell.add_command('mkdir -p %s/%s'%(out_dir,plots_dir))
        shell.run()
        shell.clean()

        plots_dir = '%s/%s'%(out_dir,plots_dir)
        time_steps = self.time_steps
        lwidth=0.8
        ts_colors = plt.cm.gist_rainbow(np.linspace(0.,1.,num=self.NSimulations))

        # Plot for each time step
        for ts in range(self.NSimulations):
        
            pol   = self.RToutput[ts].polarization
            pol_sq = pol[0]*pol[0] + pol[1]*pol[1] + pol[2]*pol[2]
            times = np.linspace(0.,self.NETime,num=pol.shape[1])
            f, (axes) = plt.subplots(4,1,sharex=True)
            axes[0].plot(times, pol[0], '-', lw=lwidth, color='blue',  label='pol-x')
            axes[1].plot(times, pol[1], '-', lw=lwidth, color='green', label='pol-y') 
            axes[2].plot(times, pol[2], '-', lw=lwidth, color='red',   label='pol-z')
            axes[3].plot(times, pol_sq, '-', lw=lwidth, color='orange',label='|pol|^2')
            for ax in axes:
                ax.axhline(0.,lw=0.5,color='gray',zorder=-5)
                ax.legend(loc='upper left')
            f.tight_layout()
             
            plt.savefig('%s/polarizations_%d%s.png'%(plots_dir,self.time_steps[ts]*odm,units),format='png',dpi=150)

        # Plot for all time steps
        f, (axes) = plt.subplots(4,1,sharex=True)
        for ts in range(self.NSimulations):

            label = '%d%s'%(time_steps[ts]*odm,units)
            pol   = self.RToutput[ts].polarization
            pol_sq = pol[0]*pol[0] + pol[1]*pol[1] + pol[2]*pol[2]
            times = np.linspace(0.,self.NETime,num=pol.shape[1])
            axes[0].plot(times, pol[0], '-', lw=lwidth, color=ts_colors[ts], label=label)
            axes[1].plot(times, pol[1], '-', lw=lwidth, color=ts_colors[ts], label=label)
            axes[2].plot(times, pol[2], '-', lw=lwidth, color=ts_colors[ts], label=label)
            axes[3].plot(times, pol_sq, '-', lw=lwidth, color=ts_colors[ts], label=label)
            handles, labels = axes[3].get_legend_handles_labels()
        for ax in axes: ax.axhline(0.,lw=0.5,color='gray',zorder=-5)
        
        f.legend(handles, labels, loc='center right')
        f.tight_layout()

        plt.savefig('%s/polarizations_comparison.png'%plots_dir,format='png',dpi=150) 

        # Plot for all time steps |pol|^2
        f, (axes) = plt.subplots(self.NSimulations,1,sharex=True)
        for ts in range(self.NSimulations):

            pol = self.RToutput[ts].polarization
            pol_sq = pol[0]*pol[0] + pol[1]*pol[1] + pol[2]*pol[2]
            times = np.linspace(0.,self.NETime,num=pol.shape[1])
            pol_ts_label = "%d%s"%(time_steps[ts]*odm,units)
            axes[ts].plot(times, pol_sq, '-', lw=lwidth, color=ts_colors[ts], label=pol_ts_label)
        for ax in axes:
            ax.axhline(0.,lw=0.5,color='gray',zorder=-5)
            ax.legend(loc='upper left')
        f.tight_layout()

        plt.savefig('%s/polarizations_squared.png'%plots_dir,format='png',dpi=150)

        # Plot for all time steps along field direction
        f, (axes) = plt.subplots(self.NSimulations,1,sharex=True)
        for ts in range(self.NSimulations):
    
            pol = self.RToutput[ts].polarization
            pol_x,pol_label = self.pol_along_field(pol)
            times = np.linspace(0.,self.NETime,num=pol.shape[1])
            pol_ts_label = "%s_%d%s"%(pol_label,time_steps[ts]*odm,units)
            axes[ts].plot(times, pol_x, '-', lw=lwidth, color=ts_colors[ts], label=pol_ts_label) 
        for ax in axes:
            ax.axhline(0.,lw=0.5,color='gray',zorder=-5)
            ax.legend(loc='upper left')
        f.tight_layout()        

        plt.savefig('%s/polarizations_field_direction.png'%plots_dir,format='png',dpi=150)

    # OLD_VERSION: IN-WORKFLOW wait_for_job
    # def wait_for_job(self,shell,time_step=10.):
    #     """
    #     Let the python execution sleep until job completion
    #     """
    #     job_status = shell.check_job_status(self.RUN_path)
    #     condition = job_status=='R' or job_status=='PD' or job_status=='CG'
    #     while condition:
    #         time.sleep(time_step)
    #         job_status = shell.check_job_status(self.RUN_path) 
    #         condition = job_status=='R' or job_status=='PD' or job_status=='CG'



"""
GNU GPLv2 License (Abipy)

abipy is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2 as published by
the Free Software Foundation

abipy is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with abipy.  
If not, see http://www.gnu.org/licenses/.

SKW interpolation: 

Shankland-Koelling-Wood Fourier interpolation scheme.
For the theoretical background see :cite:`Euwema1969,Koelling1986,Pickett1988,Madsen2006`.
"""
#This class is imported from the abipy package: https://github.com/abinit/abipy

import itertools
import numpy as np
import scipy
import time
from collections import deque
from monty.termcolor import cprint
from monty.collections import dict2namedtuple
from yambopy.tools.citations import citation

class SkwInterpolator():
    """
    This object implements the Shankland-Koelling-Wood Fourier interpolation scheme.
    It can be used to interpolate functions in k-space with the periodicity of the
    reciprocal lattice and satisfying F(k) = F(Sk) for each rotation S
    belonging to the point group of the crystal. For readability reason,
    the names of the variables are chosen assuming we are interpolating electronic eigenvalues
    but the same object can be used to interpolate other quantities. Just set the first dimension to 1.
    """

    @citation("SKW interpolation from AbiPy: X. Gonze et al., Comput. Phys. Commun. 248, 107042 (2020)")
    def __init__(self, lpratio, kpts, eigens, fermie, nelect, cell, symrel, has_timrev,
                 filter_params=None, verbose=1):
        """
        Args:
            lpratio: Ratio between the number of star-functions and the number of ab-initio k-points.
                5-10 should be OK in many systems, larger values may be required for accurate derivatives.
            kpts: numpy array with the [nkpt, 3] ab-initio k-points in reduced coordinates.
            eigens: numpy array with the ab-initio energies. shape [nsppol, nkpt, nband].
            fermie: Fermi energy in eV.
            nelect: Number of electrons in the unit cell
            cell: (lattice, positions, numbers)
                lattice: numpy array with direct lattice vectors along the rows.
                positions: atomic positions in reduced coordinates.
                numbers: Atomic number for each atom in positions.
            symrel: [nsym, 3, 3] numpy array with the (ferromagnetic) symmetry operations of the direct lattice
                in reduced coordinates. anti-ferromagnetic symmetries (if any) should be removed by the caller.
            has_timrev: True is time-reversal can be used.
            filter_params: List with parameters used to filter high-frequency components (Eq 9 of PhysRevB.61.1639)
                First item gives rcut, second item sigma. Ignored if None.
            verbose: Verbosity level.
        """
        self.verbose = verbose
        self.cell = cell
        lattice = self.cell[0]
        self.original_fermie = fermie
        self.interpolated_fermie = self.original_fermie
        self.nelect = nelect
        self.has_timrev = has_timrev

        # iscomplexobj is used to handle lifetimes.
        eigens = np.atleast_3d(eigens)
        self.iscomplexobj = np.iscomplexobj(eigens)
        self.nsppol, self.nkpt, self.nband = eigens.shape

        if len(kpts) != self.nkpt:
            raise ValueError("Second dimension of eigens should be %d but got array of shape: %s" %
                (len(kpts), eigens.shape))
        if self.nkpt == 1:
            raise ValueError("Interpolation algorithm requires nkpt > 1")

        rprimd = np.asarray(lattice).T
        self.rmet = np.matmul(rprimd.T, rprimd)

        # Find point group operations.
        symrel = np.reshape(symrel, (-1, 3, 3))
        self.ptg_symrel, self.ptg_symrec, has_inversion = extract_point_group(symrel, has_timrev)
        self.ptg_nsym = len(self.ptg_symrel)
        if self.verbose:
            print("Found", self.ptg_nsym, "symmetries in point group")

        # Find nrwant star points.
        self.lpratio = lpratio = int(lpratio)
        if lpratio <= 1:
            raise ValueError("lpratio must be > 1 but got %s" % lpratio)

        nrwant = lpratio * self.nkpt
        fact = 1/2 if has_inversion else 1
        rmax = int((1.0 + (lpratio * self.nkpt * self.ptg_nsym * fact) / 2.0) ** (1/3.)) * np.ones(3, dtype=int)
        #rmax = int((1.0 + (lpratio * self.nkpt) / 2.0) ** (1/3.)) * np.ones(3, dtype=int)

        while True:
            self.rpts, r2vals, ok = self._find_rstar_gen(nrwant, rmax)
            self.nr = len(self.rpts)
            if ok:
                break
            else:
                print("rmax: ", rmax," was not large enough to find", nrwant, "R-star points.")
                rmax *= 2
                print("Will try again with enlarged rmax:", rmax)

        print("Using:", self.nr, "star-functions. nstars/nk:", self.nr / self.nkpt)

        # Compute (inverse) roughness function.
        c1, c2 = 0.25, 0.25
        r2min = r2vals[1]
        inv_rhor = 1.0 / ((1.0 - c1 * r2vals / r2min) ** 2 + c2 * (r2vals / r2min) ** 3)

        # Construct star functions for the ab-initio k-points.
        nsppol, nband, nkpt, nr = self.nsppol, self.nband, self.nkpt, self.nr
        self.skr = np.empty((nkpt, nr), dtype=complex)
        for ik, kpt in enumerate(kpts):
            self.skr[ik] = self.get_stark(kpt)

        # Build H(k,k') matrix (Hermitian)
        hmat = np.empty((nkpt-1, nkpt-1), dtype=complex)
        for jk in range(nkpt-1):
            v_jkr = self.skr[jk, 1:] - self.skr[nkpt-1, 1:]
            #for ik in range(jk + 1):
            for ik in range(nkpt-1):
                v_ikr = inv_rhor[1:] * (self.skr[ik, 1:] - self.skr[nkpt-1, 1:])
                hmat[ik, jk] = np.vdot(v_jkr, v_ikr)
                if ik == jk: hmat[ik, jk] = hmat[ik, jk].real

        # Solving system of linear equations to get lambda coeffients (eq. 10 of PRB 38 2721)..."
        de_kbs = np.empty((nkpt-1, nband, nsppol), dtype=complex)
        for spin in range(nsppol):
            for ib in range(nband):
                de_kbs[:, ib, spin] = eigens[spin, 0:nkpt-1, ib] - eigens[spin, nkpt-1, ib]

        # Solve all bands and spins at once
        # FIXME: Portability problem with scipy 0.19 in which linalg.solve wraps the expert drivers
        # http://scipy.github.io/devdocs/release.0.19.0.html#foreign-function-interface-improvements
        if scipy.__version__ == "0.19.0":
            import warnings
            warnings.warn("linalg.solve in scipy 0.19.0 gives weird results. Use at your own risk!!!")

        try:
            lmb_kbs = scipy.linalg.solve(hmat, np.reshape(de_kbs, (-1, nband * nsppol)))
            #lmb_kbs = scipy.linalg.solve(hmat, np.reshape(de_kbs, (-1, nband * nsppol)),
                    #sym_pos=True, lower=False, overwrite_a=True, overwrite_b=True, check_finite=False)
                    #sym_pos=False, lower=False, overwrite_a=False, overwrite_b=False, check_finite=True)

        except scipy.linalg.LinAlgError as exc:
            print("Cannot solve system of linear equations to get lambda coeffients (eq. 10 of PRB 38 2721)")
            print("This usually happens when there are symmetrical k-points passed to the interpolator.")
            raise exc

        lmb_kbs = np.reshape(lmb_kbs, (-1, nband, nsppol))

        # Compute coefficients.
        self.coefs = np.empty((nsppol, nband, nr), dtype=complex)
        for spin in range(nsppol):
            for ib in range(nband):
                for ir in range(1, nr):
                    self.coefs[spin, ib, ir] = inv_rhor[ir] \
                        * np.vdot(self.skr[:nkpt-1, ir] - self.skr[nkpt-1, ir], lmb_kbs[:nkpt-1, ib, spin])

                self.coefs[spin, ib, 0] = eigens[spin, nkpt-1, ib] \
                    - np.dot(self.coefs[spin, ib, 1:nr], self.skr[nkpt-1, 1:nr])

        # Filter high-frequency.
        self.rcut, self.rsigma = None, None
        if filter_params is not None:
            self.rcut = filter_params[0] * np.sqrt(r2vals[-1])
            self.rsigma = rsigma = filter_params[1]
            if self.verbose:
                print("Applying filter (Eq 9 of PhysRevB.61.1639) with rcut:", self.rcut, ", rsigma", self.rsigma)
            from scipy.special import erfc
            for ir in range(1, nr):
                self.coefs[:, :, ir] *= 0.5 * erfc((np.sqrt(r2vals[ir]) - self.rcut) / self.rsigma)

        # Prepare workspace arrays for star functions.
        self.cached_kpt = np.ones(3) * np.inf
        self.cached_kpt_dk1 = np.ones(3) * np.inf
        self.cached_kpt_dk2 = np.ones(3) * np.inf

        # Compare ab-initio data with interpolated results.
        mae = 0.0
        for spin in range(nsppol):
            for ik, kpt in enumerate(kpts):
                skw_eb = self.eval_sk(spin, kpt)
                mae += np.abs(eigens[spin, ik] - skw_eb).sum()
                if self.verbose >= 10:
                    # print interpolated eigenvales
                    for band in range(self.nband):
                        e0 = eigens[spin, ik, band]
                        eskw = skw_eb[band]
                        print("spin", spin, "band", band, "ikpt", ik, "e0", e0, "eskw", eskw, "diff", e0 - eskw)

        mae *= 1e3 / (nsppol * nkpt * nband)
        if np.isnan(mae) or np.isinf(mae) or mae > 1000:
            raise RuntimeError("Interpolation went bananas! mae = %s" % mae)

        warn = mae > 10.0
        cprint("FIT vs input data: Mean Absolute Error= %.3e (meV)" % mae, color="red" if warn else "green")
        if warn:
            # Issue warning if error too large.
            cprint("Large error in SKW interpolation!", "red")
            cprint("MAE: %.3e (meV)" % mae, color= "red")

        self.mae = mae

    def __str__(self):
        return self.to_string()

    def to_string(self, verbose: int = 0) -> str:
        """String representation."""
        lines = []
        app = lines.append
        app("nsppol: %s, nband: %s" % (self.nsppol, self.nband))
        app("Number of operations in point-group: %s with time-reversal: %s" % (self.ptg_nsym, self.has_timrev))
        app("Number of ab-initio k-points: %s" % self.nkpt)
        app("Number of star functions: %s [nstars/nk = %s]" % (self.nr, (self.nr / self.nkpt)))
        if self.rcut is not None:
            app("Fourier filter (Eq 9 of PhysRevB.61.1639) with rcut: %s, rsigma: %s" % (self.rcut, self.rsigma))
        app("Comparison between ab-initio data and fit gave Mean Absolute Error: %s [meV]" % self.mae)

        return "\n".join(lines)

    def interp_kpts(self, kfrac_coords, dk1=False, dk2=False):
        """
        Interpolate energies on an arbitrary set of k-points. Optionally, compute
        gradients and Hessian matrices.

        Args:
            kfrac_coords: K-points in reduced coordinates.
            dk1 (bool): True if gradient is wanted.
            dk2 (bool): True to compute 2nd order derivatives.

        Return:
            namedtuple with:
            interpolated energies in eigens[nsppol, len(kfrac_coords), nband]
            gradient in dedk[self.nsppol, len(kfrac_coords), self.nband, 3))
            hessian in dedk2[self.nsppol, len(kfrac_coords), self.nband, 3, 3))

            gradient and hessian are set to None if not computed.
        """
        start = time.time()

        kfrac_coords = np.reshape(kfrac_coords, (-1, 3))
        new_nkpt = len(kfrac_coords)
        new_eigens = np.empty((self.nsppol, new_nkpt, self.nband))

        dedk = None if not dk1 else np.empty((self.nsppol, new_nkpt, self.nband, 3))
        dedk2 = None if not dk2 else np.empty((self.nsppol, new_nkpt, self.nband, 3, 3))

        der1, der2 = None, None
        for spin in range(self.nsppol):
            for ik, newk in enumerate(kfrac_coords):
                if dk1: der1 = dedk[spin, ik]
                if dk2: der2 = dedk2[spin, ik]
                new_eigens[spin, ik] = self.eval_sk(spin, newk, der1=der1, der2=der2)

        if self.verbose:
            print("Interpolation completed in %.3f (s)" % (time.time() - start))

        return dict2namedtuple(eigens=new_eigens, dedk=dedk, dedk2=dedk2)

    def interp_kpts_and_enforce_degs(self, kfrac_coords, ref_eigens, atol=1e-4):
        """
        Interpolate energies on an arbitrary set of k-points. Use `ref_eigens`
        to detect degeneracies and average the interpolated values in the degenerate subspace.
        """
        kfrac_coords = np.reshape(kfrac_coords, (-1, 3))
        new_nkpt = len(kfrac_coords)
        ref_eigens = np.reshape(ref_eigens, (self.nsppol, new_nkpt, self.nband))

        # Interpolate eigenvales.
        new_eigens = self.interp_kpts(kfrac_coords).eigens

        # Average interpolated values over degenerates bands.
        for spin in range(self.nsppol):
            for ik in range(new_nkpt):
                for dgbs in find_degs_sk(ref_eigens[spin, ik], atol):
                    if len(dgbs) == 1: continue
                    new_eigens[spin, ik, dgbs] = new_eigens[spin, ik, dgbs].sum() / len(dgbs)

        return dict2namedtuple(eigens=new_eigens, dedk=None, dedk2=None)

    def eval_sk(self, spin, kpt, der1=None, der2=None) -> np.ndarray:
        """
        Interpolate eigenvalues for all bands at a given (spin, k-point).
        Optionally compute gradients and Hessian matrices.

        Args:
            spin: Spin index.
            kpt: K-point in reduced coordinates.
            der1: If not None, ouput gradient is stored in der1[nband, 3].
            der2: If not None, output Hessian is der2[nband, 3, 3].

        Return:
            oeigs[nband]
        """
        # [NB, NR] x [NR]
        oeigs = np.matmul(self.coefs[spin], self.get_stark(kpt))
        if not self.iscomplexobj: oeigs = oeigs.real

        if der1 is not None:
            skr_dk1 = self.get_stark_dk1(kpt)
            for ii in range(3):
                value = np.matmul(self.coefs[spin, :, :], skr_dk1[ii])
                if not self.iscomplexobj: value = value.real
                der1[:, ii] = value

        if der2 is not None:
            skr_dk2 = self.get_stark_dk2(kpt)
            for jj in range(3):
                for ii in range(jj + 1):
                    value = np.matmul(self.coefs[spin, :, :], skr_dk2[ii,jj])
                    if not self.iscomplexobj: value = value.real
                    der2[:, ii, jj] = value
                    if ii != jj: der2[jj, ii] = der2[ii, jj]

        return oeigs

    #def eval_skb(self, spin, kpt, band, der1=None, der2=None):
    #    """
    #    Interpolate eigenvalues for a given (spin, k-point, band).

    #    Args:

    #    Return:
    #    """
    #    # Compute star function for this k-point (if not already in memory)
    #    if np.any(kpt != self.cached_kpt):
    #        self.cached_skr, self.cached_kpt = self.get_stark(kpt), kpt

    #    # Do not take complex conjugate.
    #    oeig = np.dot(self.coefs[spin, band], self.cached_skr)
    #    if not self.iscomplexobj: oeigs = oeigs.real
    #    if der1 is None and der2 is None: return oeig

    #    # TODO: Test Derivatives
    #    if der1 is not None:
    #        # Compute first-order derivatives.
    #        if np.any(kpt != self.cached_kpt_dk1):
    #            self.cached_skr_dk1, self.cached_kpt_dk1 = self.get_stark_dk1(kpt), kpt

    #        for ii in range(3):
    #            value = np.dot(self.coefs[spin, band], self.cached_skr_dk1[ii])
    #            if not self.iscomplexobj: value = value.real
    #            der1[ii] = value

    #        if der2 is None: return oeig, der1

    #    if der2 is not None:
    #        # Compute second-order derivatives.
    #        if np.any(kpt != self.cached_kpt_dk2):
    #            self.cached_skr_dk2, self.cached_kpt_dk2 = self.get_stark_dk2(kpt), kpt

    #        der2 = zero
    #        for jj in range(3):
    #            for ii in range(jj + 1):
    #                value = np.dot(self.coefs[spin, band], self.cached_skr_dk2[ii,jj])
    #                if not self.iscomplexobj: value = value.real
    #                der2[ii, jj] = value
    #                if ii != jj: der2[jj, ii] = der2[ii, jj]

    #    return oeig, der1, der2

    def get_stark(self, kpt) -> np.ndarray:
        """
        Return the star function for k-point `kpt`.

        Args:
            kpt: K-point in reduced coordinates.

        Return:
            complex array of shape [self.nr]
        """
        two_pi = 2.0 * np.pi
        skr = np.zeros(self.nr, dtype=complex)
        _np_exp = np.exp
        for omat in self.ptg_symrel:
            sk = two_pi * np.matmul(omat.T, kpt)
            skr += _np_exp(1.j * np.matmul(self.rpts, sk))
        skr /= self.ptg_nsym

        return skr

    def get_stark_dk1(self, kpt) -> np.ndarray:
        """
        Compute the 1st-order derivative of the star function wrt k

        Args:
            kpt: K-point in reduced coordinates.

        Return:
            complex array [3, self.nr]  with the derivative of the
            star function wrt k in reduced coordinates.
        """
        srk_dk1 = np.zeros((3, self.nr), dtype=complex)
        two_pi = 2.0 * np.pi
        rpts_t = self.rpts.T

        for omat in self.ptg_symrel:
            sk = two_pi * np.matmul(omat.T, kpt)
            exp_skr = np.exp(1.j * np.matmul(self.rpts, sk))
            for ir, rr in enumerate(self.rpts):
                srk_dk1[:, ir] += exp_skr[ir] * np.matmul(omat, rr)
            #for ir, or in enumerate(np.matmul(omat, rpts_t).T):
            #    srk_dk1[:, ir] += exp_skr[ir] * or

        srk_dk1 *= 1.j / self.ptg_nsym
        return srk_dk1

    def get_stark_dk2(self, kpt) -> np.ndarray:
        """
        Compute the 2nd-order derivatives of the star function wrt k.

        Args:
            kpt: K-point in reduced coordinates.

        Return:
            Complex numpy array of shape [3, 3, self.nr] with the 2nd-order derivatives
            of the star function wrt k in reduced coordinates.
        """
        srk_dk2 = np.zeros((3, 3, self.nr), dtype=complex)
        raise NotImplementedError()
        #work = zero
        #do isym=1,self.ptg_nsym
        #   sk = two_pi * matmul(transpose(self%ptg_symrel(:,:,isym)), kpt)
        #   do ir=1,self%nr
        #     sr = matmul(self%ptg_symrel(:,:,isym), self%rpts(:,ir))
        #     eiskr = exp(j_dpc * dot_product(sk, self%rpts(:,ir)))
        #     do jj=1,3
        #        do ii=1,jj
        #            work(ii,jj,ir) = work(ii,jj,ir) + eiskr * sr(ii) * sr(jj)

        #work = - work / self.ptg_nsym

        #do jj=1,3
        #   do ii=1,jj
        #       srk_dk2(:, ii, jj) = work(ii, jj, :)
        #       if (ii /= jj) srk_dk2(:,jj,ii) = work(:,ii,jj)

        return srk_dk2

    #def find_stationary_points(self, kmesh, bstart=None, bstop=None, is_shift=None)
    #    k = self.get_sampling(kmesh, is_shift)
    #    if bstart is None: bstart = self.nelect // 2 - 1
    #    if bstop is None: bstop = self.nelect // 2
    #    nb = bstop - bstart + 1
    #    results = []
    #    for ik_ibz, kpt in enumerate(k.ibz):
    #        vk_b = self.eval_dk1(kpt, bstart, bstop)
    #        bands = []
    #        for ib, v in enumerate(vk_b):
    #            if v < atol: bands.append(ib + bstart)
    #        if bands:
    #            #results.append()
    #            for band in bands:
    #                d2k_b = self.eval_dk2(kpt, band)

    #    return results

    def _find_rstar_gen(self, nrwant, rmax) -> tuple:
        """
        Find all lattice points generating the stars inside the supercell defined by `rmax`

        Args:
            nrwant: Number of star-functions required.
            rmax: numpy array with the maximum number of cells along the 3 reduced directions.

        Returns:
            tuple: (rpts, r2vals, ok)
        """
        msize = (2 * rmax + 1).prod()
        rtmp = np.empty((msize, 3), dtype=int)
        r2tmp = np.empty(msize)
        if self.verbose: print("rmax", rmax, "msize:", msize)

        start = time.time()
        for cnt, l in enumerate(itertools.product(range(-rmax[0], rmax[0] + 1),
                                                  range(-rmax[1], rmax[1] + 1),
                                                  range(-rmax[2], rmax[2] + 1))):
            rtmp[cnt] = l
            r2tmp[cnt] = np.dot(l, np.matmul(self.rmet, l))

        if self.verbose: print("gen points", time.time() - start)

        start = time.time()
        # Sort r2tmp and rtmp
        iperm = np.argsort(r2tmp)
        r2tmp = r2tmp[iperm]
        rtmp = rtmp[iperm]
        #return rtmp, r2tmp, True

        # Find shells
        r2sh = np.empty(msize, dtype=int)    # Correspondence between R and shell index.
        shlim = np.empty(msize, dtype=int)   # For each shell, the index of the initial G-vector.
        nsh = 1
        r2sh[0] = 0
        shlim[0] = 0
        r2_prev = 0.0
        for ir in range(1, msize):
            if abs(r2tmp[ir] - r2_prev) > r2tmp[ir] * 1e-8:
                r2_prev = r2tmp[ir]
                shlim[nsh] = ir
                nsh += 1
            r2sh[ir] = nsh
        shlim[nsh] = msize
        if self.verbose:
            print("nshells", nsh)
            print("shells", time.time() - start)

        # Find R-points generating the stars.
        # This is the most CPU critical part. I think it's difficult to do better than this without cython
        start = time.time()
        rgen = deque()

        for ish in range(nsh):
            ss, ee = shlim[ish], shlim[ish + 1]
            if ss + 1 == ee:
                rgen.append(rtmp[ss])
                continue

            if True:
                # This is faster
                rs = set([tuple(rtmp[ss])])
                for ir in range(ss + 1, ee):
                    rvec = rtmp[ir]
                    if all(tuple(np.matmul(rot, rvec)) not in rs for rot in self.ptg_symrel):
                        rs.add(tuple(rvec))
            else:
                rs = deque([rtmp[ss]])
                for ir in range(ss + 1, ee):
                    for rot in self.ptg_symrel:
                        rot_r = np.matmul(rot, rtmp[ir])
                        if any(np.all(rot_r == x) for x in rs): break
                    else:
                        # We have new R-point.
                        #print("Adding new point")
                        rs.append(rtmp[ir])

            #print(len(rs), rs)
            rgen.extend(rs)
        #print(rgen)
        if self.verbose: print("stars", time.time() - start)

        start = time.time()
        rgen = np.array(rgen, dtype=int)
        nstars = len(rgen)

        # Store rpts and compute ||R||**2.
        ok = nstars >= nrwant
        nr = min(nstars, nrwant)
        rpts = rgen[:nr].copy()
        r2vals = np.empty(nr)
        for ir in range(nr):
            r2vals[ir] = np.dot(rpts[ir], np.matmul(self.rmet, rpts[ir]))

        if self.verbose:
            print("r2max ", rpts[nr-1])
            print("end ", time.time() - start)
            if self.verbose > 10:
                print("nstars:", nstars)
                for r, r2 in zip(rpts, r2vals):
                    print(r, r2)

        return rpts, r2vals, ok


def extract_point_group(symrel, has_timrev) -> tuple:
    """
    Extract the point group rotations from the spacegroup. Add time-reversal
    if spatial inversion is not present and `has_timrev`.

    Return:
        (ptg_symrel, ptg_symrec) with rotations in real- and reciprocal-space.
    """
    nsym = len(symrel)
    tmp_nsym = 1
    work_symrel = np.empty((2*nsym, 3, 3), dtype=int)
    work_symrel[0] = symrel[0]

    for isym in range(1, nsym):
        found = any(np.all(symrel[isym] == work_symrel[search]) for search in range(tmp_nsym))
        if not found:
            work_symrel[tmp_nsym] = symrel[isym]
            tmp_nsym += 1

    inversion_3d = -np.eye(3, dtype=int)
    has_inversion = any(np.all(w == inversion_3d) for w in work_symrel[:tmp_nsym])

    # Now we know the symmetries of the point group.
    ptg_nsym = 2 * tmp_nsym if not has_inversion and has_timrev else tmp_nsym
    ptg_symrel = np.empty((ptg_nsym, 3, 3), dtype=int)
    ptg_symrec = np.empty((ptg_nsym, 3, 3), dtype=int)

    ptg_symrel[:tmp_nsym] = work_symrel[:tmp_nsym]
    for isym in range(tmp_nsym):
        ptg_symrec[isym] = np.linalg.inv(ptg_symrel[isym]).T

    if not has_inversion and has_timrev:
        # Add inversion.
        ptg_symrel[tmp_nsym:] = -work_symrel[:tmp_nsym]
        for isym in range(tmp_nsym, ptg_nsym):
            ptg_symrec[isym] = np.linalg.inv(ptg_symrel[isym]).T

    return ptg_symrel, ptg_symrec, has_inversion



# Copyright (C) 2018 Henrique Pereira Coutada Miranda, Alejandro Molina Sanchez
# All rights reserved.
#
# This file is part of yambopy
#
from __future__ import print_function
import os

__all__ = ['ProjwfcIn']

class ProjwfcIn():
    """
    A class to generate and manipulate projwfc input files.
    """    
    _projwfc = 'projwfc.x'

    def __init__(self,prefix,DeltaE = 0.05,folder='.'):
        self.prefix = prefix
        self.DeltaE = DeltaE
        
    def run(self,filename=None,procs=1,folder='.'):
        """ 
        Function is used to run this job locally
        """
        filename = self.getfilename(filename)
        os.system('mkdir -p %s'%folder)
        self.write("%s/%s"%(folder,filename))
        print("Run projwfc...")
        if procs == 1:
            os.system('cd %s; %s -inp %s > projwfc.log' % (folder,self._projwfc,filename))
        else:
            os.system('cd %s; mpirun -np %d %s -inp %s > projwfc.log' % (folder,procs,self._projwfc,filename))
        print("done!")

    def getfilename(self,filename):
        if filename is None:
            filename = "%s.projwfc"%self.prefix
        return filename

    def write(self,filename=None,folder='.'):
        """
        Write the projwfc file
        """
        filename = self.getfilename(filename)
        os.system('mkdir -p %s'%folder)
        f = open("%s/%s"%(folder,filename),'w')
        f.write(str(self))
        f.close()

    def __str__(self):
        """
        Output the file in the from of a string
        """
        string = '\n'
        string += "&projwfc\n"
        string += "prefix = '%s'\n"%self.prefix
        string += "DeltaE = %lf\n"%self.DeltaE
        string += "/\n"
        return string
